# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""Classes to manage the GUI presentation of installing an Addon (e.g. the sequence of dialog boxes
that do dependency resolution, error handling, etc.). See AddonInstallerGUI and MacroInstallerGUI
classes for details."""

import os
import sys
from typing import List

import FreeCAD
import FreeCADGui
from PySide import QtCore, QtWidgets

from addonmanager_installer import AddonInstaller, MacroInstaller
from addonmanager_dependency_installer import DependencyInstaller
import addonmanager_utilities as utils
from Addon import Addon, MissingDependencies

translate = FreeCAD.Qt.translate
from PySide.QtCore import QT_TRANSLATE_NOOP

# pylint: disable=c-extension-no-member,too-few-public-methods,too-many-instance-attributes


class AddonInstallerGUI(QtCore.QObject):
    """GUI functions (sequence of dialog boxes) for installing an addon interactively. The actual
    installation is handled by the AddonInstaller class running in a separate QThread. An instance
    of this AddonInstallerGUI class should NOT be run in a separate thread, but on the main GUI
    thread. All dialogs are modal."""

    # External classes are expected to "set and forget" this class, but in the event that some
    # action must be taken if the addon is actually installed, this signal is provided. Note that
    # this class already provides a "Successful installation" dialog, so external code need not
    # do so.
    success = QtCore.Signal(object)

    # Emitted once all work has been completed, regardless of success or failure
    finished = QtCore.Signal()

    def __init__(self, addon: Addon, addons: List[Addon] = None):
        super().__init__()
        self.addon_to_install = addon
        self.addons = [] if addons is None else addons
        self.installer = AddonInstaller(addon)
        self.dependency_installer = None
        self.install_worker = None
        self.dependency_dialog = None
        self.dependency_worker_thread = None
        self.dependency_installation_dialog = None
        self.installing_dialog = None
        self.worker_thread = None

        # Set up the installer connections
        self.installer.success.connect(self._installation_succeeded)
        self.installer.failure.connect(self._installation_failed)

    def __del__(self):
        if self.worker_thread and hasattr(self.worker_thread, "quit"):
            self.worker_thread.quit()
            self.worker_thread.wait(500)
            if self.worker_thread.isRunning():
                FreeCAD.Console.PrintError(
                    "INTERNAL ERROR: Thread did not quit() cleanly, using terminate()\n"
                )
                self.worker_thread.terminate()

    def run(self):
        """Instructs this class to begin displaying the necessary dialogs to guide a user through
        an Addon installation sequence. All dialogs are modal."""

        # Dependency check
        deps = MissingDependencies(self.addon_to_install, self.addons)

        # Python interpreter version check
        stop_installation = self._check_python_version(deps)
        if stop_installation:
            self.finished.emit()
            return

        # Required Python
        if hasattr(deps, "python_requires") and deps.python_requires:
            # Disallowed packages:
            stop_installation = self._handle_disallowed_python(deps.python_requires)
            if stop_installation:
                self.finished.emit()
                return
            # Allowed but uninstalled is handled below

        # Remove any disallowed packages from the optional list
        if hasattr(deps, "python_optional") and deps.python_optional:
            self._clean_up_optional(deps)

        # Missing FreeCAD workbenches
        if hasattr(deps, "wbs") and deps.wbs:
            stop_installation = self._report_missing_workbenches(deps.wbs)
            if stop_installation:
                self.finished.emit()
                return

        # If we have any missing dependencies, display a dialog to the user asking if they want to
        # install them.
        if deps.external_addons or deps.python_requires or deps.python_optional:
            # Recoverable: ask the user if they want to install the missing deps, do so, then
            # proceed with the installation
            self._resolve_dependencies_then_install(deps)
        else:
            # No missing deps, just install
            self.install()

    def _handle_disallowed_python(self, python_requires: List[str]) -> bool:
        """Determine if we are missing any required Python packages that are not in the allowed
        packages list. If so, display a message to the user, and return True. Otherwise return
        False."""

        bad_packages = []
        for dep in python_requires:
            if dep.lower() not in AddonInstaller.allowed_packages:
                bad_packages.append(dep)

        for dep in bad_packages:
            python_requires.remove(dep)

        if bad_packages:
            # pylint: disable=line-too-long
            message = (
                "<p>"
                + translate(
                    "AddonsInstaller",
                    "This addon requires Python packages that are not installed, and cannot be installed automatically. To use this workbench you must install the following Python packages manually:",
                )
                + "</p><ul>"
            )
            if len(bad_packages) < 15:
                for dep in bad_packages:
                    message += f"<li>{dep}</li>"
            else:
                message += "<li>(" + translate("AddonsInstaller", "Too many to list") + ")</li>"
            message += "</ul>"
            message += "To ignore this error and install anyway, press OK."
            r = QtWidgets.QMessageBox.critical(
                utils.get_main_am_window(),
                translate("AddonsInstaller", "Missing Requirement"),
                message,
                QtWidgets.QMessageBox.Ok | QtWidgets.QMessageBox.Cancel,
            )

            if r == QtWidgets.QMessageBox.Ok:
                # Force the installation to proceed
                return False
            return True
        return False

    def _report_missing_workbenches(self, wbs) -> bool:
        """If there are missing workbenches, display a dialog informing the user. Returns True to
        stop the installation, or False to proceed."""
        addon_name = self.addon_to_install.name
        if len(wbs) == 1:
            name = wbs[0]
            message = translate(
                "AddonsInstaller",
                "Addon '{}' requires '{}', which is not available in your copy of FreeCAD.",
            ).format(addon_name, name)
        else:
            # pylint: disable=line-too-long
            message = (
                "<p>"
                + translate(
                    "AddonsInstaller",
                    "Addon '{}' requires the following workbenches, which are not available in your copy of FreeCAD:",
                ).format(addon_name)
                + "</p><ul>"
            )
            for wb in wbs:
                message += "<li>" + wb + "</li>"
            message += "</ul>"
            message += translate("AddonsInstaller", "Press OK to install anyway.")
        r = QtWidgets.QMessageBox.critical(
            utils.get_main_am_window(),
            translate("AddonsInstaller", "Missing Requirement"),
            message,
            QtWidgets.QMessageBox.Ok | QtWidgets.QMessageBox.Cancel,
        )
        return r == QtWidgets.QMessageBox.Cancel

    def _resolve_dependencies_then_install(self, missing) -> None:
        """Ask the user how they want to handle dependencies, do that, then install."""
        self.dependency_dialog = FreeCADGui.PySideUic.loadUi(
            os.path.join(os.path.dirname(__file__), "dependency_resolution_dialog.ui")
        )
        self.dependency_dialog.setWindowFlag(QtCore.Qt.WindowStaysOnTopHint, True)

        for addon in missing.external_addons:
            self.dependency_dialog.listWidgetAddons.addItem(addon)
        for mod in missing.python_requires:
            self.dependency_dialog.listWidgetPythonRequired.addItem(mod)
        for mod in missing.python_optional:
            item = QtWidgets.QListWidgetItem(mod)
            item.setFlags(item.flags() | QtCore.Qt.ItemIsUserCheckable)
            item.setCheckState(QtCore.Qt.Unchecked)
            self.dependency_dialog.listWidgetPythonOptional.addItem(item)

        self.dependency_dialog.buttonBox.button(QtWidgets.QDialogButtonBox.Yes).clicked.connect(
            self._dependency_dialog_yes_clicked
        )
        self.dependency_dialog.buttonBox.button(QtWidgets.QDialogButtonBox.Ignore).clicked.connect(
            self._dependency_dialog_ignore_clicked
        )
        self.dependency_dialog.buttonBox.button(QtWidgets.QDialogButtonBox.Cancel).setDefault(True)
        self.dependency_dialog.exec()

    def _check_python_version(self, missing: MissingDependencies) -> bool:
        """Make sure we have a compatible Python version. Returns True to stop the installation
        or False to continue."""

        # For now only look at the minor version, since major is always Python 3
        minor_required = missing.python_min_version["minor"]
        if sys.version_info.minor < minor_required:
            # pylint: disable=line-too-long
            QtWidgets.QMessageBox.critical(
                utils.get_main_am_window(),
                translate("AddonsInstaller", "Incompatible Python version"),
                translate(
                    "AddonsInstaller",
                    "This Addon (or one if its dependencies) requires Python {}.{}, and your system is running {}.{}. Installation cancelled.",
                ).format(
                    missing.python_min_version["major"],
                    missing.python_min_version["minor"],
                    sys.version_info.major,
                    sys.version_info.minor,
                ),
                QtWidgets.QMessageBox.Cancel,
            )
            return True
        return False

    def _clean_up_optional(self, missing: MissingDependencies):
        good_packages = []
        for dep in missing.python_optional:
            if dep in self.installer.allowed_packages:
                good_packages.append(dep)
            else:
                FreeCAD.Console.PrintWarning(
                    translate(
                        "AddonsInstaller",
                        "Optional dependency on {} ignored because it is not in the allow-list",
                    ).format(dep)
                    + "\n"
                )
        missing.python_optional = good_packages

    def _dependency_dialog_yes_clicked(self) -> None:
        # Get the lists out of the dialog:
        addons = []
        for row in range(self.dependency_dialog.listWidgetAddons.count()):
            item = self.dependency_dialog.listWidgetAddons.item(row)
            name = item.text()
            for repo in self.addons:
                if repo.name == name or (
                    hasattr(repo, "display_name") and repo.display_name == name
                ):
                    addons.append(repo)

        python_requires = []
        for row in range(self.dependency_dialog.listWidgetPythonRequired.count()):
            item = self.dependency_dialog.listWidgetPythonRequired.item(row)
            python_requires.append(item.text())

        python_optional = []
        for row in range(self.dependency_dialog.listWidgetPythonOptional.count()):
            item = self.dependency_dialog.listWidgetPythonOptional.item(row)
            if item.checkState() == QtCore.Qt.Checked:
                python_optional.append(item.text())

        self._run_dependency_installer(addons, python_requires, python_optional)

    def _run_dependency_installer(self, addons, python_requires, python_optional):
        """Run the dependency installer (in a separate thread) for the given dependencies"""
        self.dependency_installer = DependencyInstaller(addons, python_requires, python_optional)
        self.dependency_installer.no_python_exe.connect(self._report_no_python_exe)
        self.dependency_installer.no_pip.connect(self._report_no_pip)
        self.dependency_installer.failure.connect(self._report_dependency_failure)
        self.dependency_installer.finished.connect(self._cleanup_dependency_worker)
        self.dependency_installer.finished.connect(self._report_dependency_success)

        self.dependency_worker_thread = QtCore.QThread(self)
        self.dependency_installer.moveToThread(self.dependency_worker_thread)
        self.dependency_worker_thread.started.connect(self.dependency_installer.run)
        self.dependency_installer.finished.connect(self.dependency_worker_thread.quit)

        self.dependency_installation_dialog = QtWidgets.QMessageBox(
            QtWidgets.QMessageBox.Information,
            translate("AddonsInstaller", "Installing dependencies"),
            translate("AddonsInstaller", "Installing dependencies") + "...",
            QtWidgets.QMessageBox.Cancel,
            parent=utils.get_main_am_window(),
        )
        self.dependency_installation_dialog.rejected.connect(self._cancel_dependency_installation)
        self.dependency_installation_dialog.show()
        self.dependency_worker_thread.start()

    def _cleanup_dependency_worker(self) -> None:
        return
        self.dependency_worker_thread.quit()
        self.dependency_worker_thread.wait(500)
        if self.dependency_worker_thread.isRunning():
            FreeCAD.Console.PrintError(
                "INTERNAL ERROR: Thread did not quit() cleanly, using terminate()\n"
            )
            self.dependency_worker_thread.terminate()

    def _report_no_python_exe(self) -> None:
        """Callback for the dependency installer failing to locate a Python executable."""
        if self.dependency_installation_dialog is not None:
            self.dependency_installation_dialog.hide()
        # pylint: disable=line-too-long
        result = QtWidgets.QMessageBox.critical(
            utils.get_main_am_window(),
            translate("AddonsInstaller", "Cannot execute Python"),
            translate(
                "AddonsInstaller",
                "Failed to automatically locate your Python executable, or the path is set incorrectly. Please check the Addon Manager preferences setting for the path to Python.",
            )
            + "\n\n"
            + translate(
                "AddonsInstaller",
                "Dependencies could not be installed. Continue with installation of {} anyway?",
            ).format(self.addon_to_install.name),
            QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
        )
        if result == QtWidgets.QMessageBox.Yes:
            self.install()
        else:
            self.finished.emit()

    def _report_no_pip(self, command: str) -> None:
        """Callback for the dependency installer failing to access pip."""
        if self.dependency_installation_dialog is not None:
            self.dependency_installation_dialog.hide()
        # pylint: disable=line-too-long
        result = QtWidgets.QMessageBox.critical(
            utils.get_main_am_window(),
            translate("AddonsInstaller", "Cannot execute pip"),
            translate(
                "AddonsInstaller",
                "Failed to execute pip, which may be missing from your Python installation. Please ensure your system has pip installed and try again. The failed command was: ",
            )
            + f"\n\n{command}\n\n"
            + translate(
                "AddonsInstaller",
                "Continue with installation of {} anyway?",
            ).format(self.addon_to_install.name),
            QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
        )
        if result == QtWidgets.QMessageBox.Yes:
            self.install()
        else:
            self.finished.emit()

    def _report_dependency_failure(self, short_message: str, details: str) -> None:
        """Callback for dependency installation failure."""
        if self.dependency_installation_dialog is not None:
            self.dependency_installation_dialog.hide()
        if self.dependency_installer and hasattr(self.dependency_installer, "finished"):
            self.dependency_installer.finished.disconnect(self._report_dependency_success)
        FreeCAD.Console.PrintError(details + "\n")
        result = QtWidgets.QMessageBox.critical(
            utils.get_main_am_window(),
            translate("AddonsInstaller", "Package installation failed"),
            short_message
            + "\n\n"
            + translate("AddonsInstaller", "See Report View for detailed failure log.")
            + "\n\n"
            + translate(
                "AddonsInstaller",
                "Continue with installation of {} anyway?",
            ).format(self.addon_to_install.name),
            QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
        )
        if result == QtWidgets.QMessageBox.Yes:
            self.install()
        else:
            self.finished.emit()

    def _report_dependency_success(self):
        """Callback for dependency installation success."""
        if self.dependency_installation_dialog is not None:
            self.dependency_installation_dialog.hide()
        self.install()

    def _dependency_dialog_ignore_clicked(self) -> None:
        """Callback for when dependencies are ignored."""
        self.install()

    def _cancel_dependency_installation(self) -> None:
        """Cancel was clicked in the dialog. NOTE: Does no cleanup, the state after cancellation is
        unknown. In most cases pip can recover from whatever we've done to it."""
        self.dependency_worker_thread.blockSignals(True)
        self.dependency_installer.blockSignals(True)
        self.dependency_worker_thread.requestInterruption()
        self.dependency_installation_dialog.hide()
        self.dependency_worker_thread.wait()
        self.finished.emit()

    def install(self) -> None:
        """Installs or updates a workbench, macro, or package"""
        self.worker_thread = QtCore.QThread()
        self.installer.moveToThread(self.worker_thread)
        self.installer.finished.connect(self.worker_thread.quit)
        self.worker_thread.started.connect(self.installer.run)

        self.installing_dialog = QtWidgets.QMessageBox(
            QtWidgets.QMessageBox.NoIcon,
            translate("AddonsInstaller", "Installing Addon"),
            translate("AddonsInstaller", "Installing FreeCAD Addon '{}'").format(
                self.addon_to_install.display_name
            ),
            QtWidgets.QMessageBox.Cancel,
            parent=utils.get_main_am_window(),
        )
        self.installing_dialog.rejected.connect(self._cancel_addon_installation)
        self.installer.finished.connect(self.installing_dialog.hide)
        self.installing_dialog.show()
        self.worker_thread.start()  # Returns immediately

    def _cancel_addon_installation(self):
        dlg = QtWidgets.QMessageBox(
            QtWidgets.QMessageBox.NoIcon,
            translate("AddonsInstaller", "Cancelling"),
            translate("AddonsInstaller", "Cancelling installation of '{}'").format(
                self.addon_to_install.display_name
            ),
            QtWidgets.QMessageBox.NoButton,
            parent=utils.get_main_am_window(),
        )
        dlg.show()
        if self.worker_thread.isRunning():
            # Interruption can take a second or more, depending on what was being done. Make sure
            # we stay responsive and update the dialog with the text above, etc.
            self.worker_thread.requestInterruption()
            self.worker_thread.quit()
            while self.worker_thread.isRunning():
                self.worker_thread.wait(50)
                QtCore.QCoreApplication.processEvents(QtCore.QEventLoop.AllEvents)
        path = os.path.join(self.installer.installation_path, self.addon_to_install.name)
        if os.path.exists(path):
            utils.rmdir(path)
        dlg.hide()
        self.finished.emit()

    def _installation_succeeded(self):
        """Called if the installation was successful."""
        QtWidgets.QMessageBox.information(
            utils.get_main_am_window(),
            translate("AddonsInstaller", "Success"),
            translate("AddonsInstaller", "{} was installed successfully").format(
                self.addon_to_install.name
            ),
            QtWidgets.QMessageBox.Ok,
        )
        self.success.emit(self.addon_to_install)
        self.finished.emit()

    def _installation_failed(self, addon, message):
        """Called if the installation failed."""
        QtWidgets.QMessageBox.critical(
            utils.get_main_am_window(),
            translate("AddonsInstaller", "Installation Failed"),
            translate("AddonsInstaller", "Failed to install {}").format(addon.name)
            + "\n"
            + message,
            QtWidgets.QMessageBox.Cancel,
        )
        self.finished.emit()


class MacroInstallerGUI(QtCore.QObject):
    """Install a macro, providing feedback about the process via dialog boxes, and then offer to
    add the macro to a custom toolbar. Should be run on the main GUI thread: this class internally
    launches a QThread for the actual installation process."""

    # Only success should matter to external code: all user interaction is handled via this class
    success = QtCore.Signal(object)

    # Emitted once all work has been completed, regardless of success or failure
    finished = QtCore.Signal()

    def __init__(self, addon: object):
        """The provided addon object must have an attribute called "macro", and that attribute must
        itself provide a callable "install" method that takes a single string, the path to the
        installation location."""
        super().__init__()
        self.addon_to_install = addon
        self.worker_thread = None
        self.installer = MacroInstaller(self.addon_to_install)
        self.addon_params = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Addons")
        self.toolbar_params = FreeCAD.ParamGet("User parameter:BaseApp/Workbench/Global/Toolbar")
        self.macro_dir = FreeCAD.getUserMacroDir(True)

    def __del__(self):
        if self.worker_thread and hasattr(self.worker_thread, "quit"):
            self.worker_thread.quit()
            self.worker_thread.wait(500)
            if self.worker_thread.isRunning():
                FreeCAD.Console.PrintError(
                    "INTERNAL ERROR: Thread did not quit() cleanly, using terminate()\n"
                )
                self.worker_thread.terminate()

    def run(self):
        """Perform the installation, including any necessary user interaction via modal dialog
        boxes. If installation proceeds successfully to completion, emits the success() signal."""

        self.worker_thread = QtCore.QThread()
        self.installer.moveToThread(self.worker_thread)
        self.installer.finished.connect(self.worker_thread.quit)
        self.installer.success.connect(self._base_installation_success)
        self.worker_thread.started.connect(self.installer.run)
        self.worker_thread.start()  # Returns immediately

    def _base_installation_success(self):
        """Callback for a successful basic macro installation."""
        self.success.emit(self.addon_to_install)
        self._ask_to_install_toolbar_button()  # Synchronous set of modals
        self.finished.emit()

    def _ask_to_install_toolbar_button(self) -> None:
        """Presents a dialog to the user asking if they want to install a toolbar button for
        a particular macro, and walks through that process if they agree to do so."""
        do_not_show_dialog = self.addon_params.GetBool("dontShowAddMacroButtonDialog", False)
        button_exists = self._macro_button_exists()
        if not do_not_show_dialog and not button_exists:
            add_toolbar_button_dialog = FreeCADGui.PySideUic.loadUi(
                os.path.join(os.path.dirname(__file__), "add_toolbar_button_dialog.ui")
            )
            add_toolbar_button_dialog.setWindowFlag(QtCore.Qt.WindowStaysOnTopHint, True)
            add_toolbar_button_dialog.buttonYes.clicked.connect(self._install_toolbar_button)
            add_toolbar_button_dialog.buttonNever.clicked.connect(
                lambda: self.addon_params.SetBool("dontShowAddMacroButtonDialog", True)
            )
            add_toolbar_button_dialog.exec()

    def _find_custom_command(self, filename):
        """Wrap calls to FreeCADGui.Command.findCustomCommand so it can be faked in testing."""
        return FreeCADGui.Command.findCustomCommand(filename)

    def _macro_button_exists(self) -> bool:
        """Returns True if a button already exists for this macro, or False if not."""
        command = self._find_custom_command(self.addon_to_install.macro.filename)
        if not command:
            return False
        toolbar_groups = self.toolbar_params.GetGroups()
        for group in toolbar_groups:
            toolbar = self.toolbar_params.GetGroup(group)
            if toolbar.GetString(command, "*") != "*":
                return True
        return False

    def _ask_for_toolbar(self, custom_toolbars) -> object:
        """Determine what toolbar to add the icon to. The first time it is called it prompts the
        user to select or create a toolbar. After that, the prompt is optional and can be configured
        via a preference. Returns the pref group for the new toolbar."""

        # In this one spot, default True: if this is the first time we got to
        # this chunk of code, we are always going to ask.
        ask = self.addon_params.GetBool("alwaysAskForToolbar", True)

        if ask:
            select_toolbar_dialog = FreeCADGui.PySideUic.loadUi(
                os.path.join(os.path.dirname(__file__), "select_toolbar_dialog.ui")
            )
            select_toolbar_dialog.setWindowFlag(QtCore.Qt.WindowStaysOnTopHint, True)

            select_toolbar_dialog.comboBox.clear()

            for group in custom_toolbars:
                ref = self.toolbar_params.GetGroup(group)
                name = ref.GetString("Name", "")
                if name:
                    select_toolbar_dialog.comboBox.addItem(name)
                else:
                    FreeCAD.Console.PrintWarning(
                        f"Custom toolbar {group} does not have a Name element\n"
                    )
            new_menubar_option_text = translate("AddonsInstaller", "Create new toolbar")
            select_toolbar_dialog.comboBox.addItem(new_menubar_option_text)

            result = select_toolbar_dialog.exec()
            if result == QtWidgets.QDialog.Accepted:
                selection = select_toolbar_dialog.comboBox.currentText()
                if select_toolbar_dialog.checkBox.checkState() == QtCore.Qt.Unchecked:
                    self.addon_params.SetBool("alwaysAskForToolbar", False)
                else:
                    self.addon_params.SetBool("alwaysAskForToolbar", True)
                if selection == new_menubar_option_text:
                    return self._create_new_custom_toolbar()
                return self._get_toolbar_with_name(selection)
            return None

        # If none of the above code returned...
        custom_toolbar_name = self.addon_params.GetString(
            "CustomToolbarName", "Auto-Created Macro Toolbar"
        )
        toolbar = self._get_toolbar_with_name(custom_toolbar_name)
        if not toolbar:
            # They told us not to ask, but then the toolbar got deleted... ask anyway!
            ask = self.addon_params.RemBool("alwaysAskForToolbar")
            return self._ask_for_toolbar(custom_toolbars)
        return toolbar

    def _get_toolbar_with_name(self, name: str) -> object:
        """Try to find a toolbar with a given name. Returns the preference group for the toolbar
        if found, or None if it does not exist."""
        custom_toolbars = self.toolbar_params.GetGroups()
        for toolbar in custom_toolbars:
            group = self.toolbar_params.GetGroup(toolbar)
            group_name = group.GetString("Name", "")
            if group_name == name:
                return group
        return None

    def _create_new_custom_toolbar(self) -> object:
        """Create a new custom toolbar and returns its preference group."""

        # We need two names: the name of the auto-created toolbar, as it will be displayed to the
        # user in various menus, and the underlying name of the toolbar group. Both must be
        # unique.

        # First, the displayed name
        custom_toolbar_name = QT_TRANSLATE_NOOP("Workbench", "Auto-Created Macro Toolbar")
        custom_toolbars = self.toolbar_params.GetGroups()
        name_taken = self._check_for_toolbar(custom_toolbar_name)
        if name_taken:
            i = 2  # Don't use (1), start at (2)
            while True:
                test_name = custom_toolbar_name + f" ({i})"
                if not self._check_for_toolbar(test_name):
                    custom_toolbar_name = test_name
                    break
                i = i + 1

        # Second, the toolbar preference group name
        i = 1
        while True:
            new_group_name = "Custom_" + str(i)
            if new_group_name not in custom_toolbars:
                break
            i = i + 1

        custom_toolbar = self.toolbar_params.GetGroup(new_group_name)
        custom_toolbar.SetString("Name", custom_toolbar_name)
        custom_toolbar.SetBool("Active", True)
        return custom_toolbar

    def _check_for_toolbar(self, toolbar_name: str) -> bool:
        """Returns True if the toolbar exists, otherwise False"""
        return self._get_toolbar_with_name(toolbar_name) is not None

    def _install_toolbar_button(self) -> None:
        """If the user has requested a toolbar button be installed, this function is called
        to continue the process and request any additional required information."""
        custom_toolbar_name = self.addon_params.GetString(
            "CustomToolbarName", "Auto-Created Macro Toolbar"
        )

        # Default to false here: if the variable hasn't been set, we don't assume
        # that we have to ask, because the simplest is to just create a new toolbar
        # and never ask at all.
        ask = self.addon_params.GetBool("alwaysAskForToolbar", False)

        # See if there is already a custom toolbar for macros:
        top_group = self.toolbar_params
        custom_toolbars = top_group.GetGroups()
        if custom_toolbars:
            # If there are already custom toolbars, see if one of them is the one we used last time
            found_toolbar = False
            for toolbar_name in custom_toolbars:
                test_toolbar = self.toolbar_params.GetGroup(toolbar_name)
                name = test_toolbar.GetString("Name", "")
                if name == custom_toolbar_name:
                    custom_toolbar = test_toolbar
                    found_toolbar = True
                    break
            if ask or not found_toolbar:
                # We have to ask the user what to do...
                custom_toolbar = self._ask_for_toolbar(custom_toolbars)
                if custom_toolbar:
                    custom_toolbar_name = custom_toolbar.GetString("Name")
                    self.addon_params.SetString("CustomToolbarName", custom_toolbar_name)
        else:
            # Create a custom toolbar
            custom_toolbar = self.toolbar_params.GetGroup("Custom_1")
            custom_toolbar.SetString("Name", custom_toolbar_name)
            custom_toolbar.SetBool("Active", True)

        if custom_toolbar:
            self._install_macro_to_toolbar(custom_toolbar)
        else:
            FreeCAD.Console.PrintMessage("In the end, no custom toolbar was set, bailing out\n")

    def _install_macro_to_toolbar(self, toolbar: object) -> None:
        """Adds an icon for the given macro to the given toolbar."""
        menuText = self.addon_to_install.display_name
        tooltipText = f"<b>{self.addon_to_install.display_name}</b>"
        if self.addon_to_install.macro.comment:
            tooltipText += f"<br/><p>{self.addon_to_install.macro.comment}</p>"
            whatsThisText = self.addon_to_install.macro.comment
        else:
            whatsThisText = translate(
                "AddonsInstaller", "A macro installed with the FreeCAD Addon Manager"
            )
        statustipText = (
            translate("AddonsInstaller", "Run", "Indicates a macro that can be 'run'")
            + " "
            + self.addon_to_install.display_name
        )
        if self.addon_to_install.macro.icon:
            if os.path.isabs(self.addon_to_install.macro.icon):
                pixmapText = os.path.normpath(self.addon_to_install.macro.icon)
            else:
                pixmapText = os.path.normpath(
                    os.path.join(self.macro_dir, self.addon_to_install.macro.icon)
                )
        elif self.addon_to_install.macro.xpm:
            icon_file = os.path.normpath(
                os.path.join(self.macro_dir, self.addon_to_install.macro.name + "_icon.xpm")
            )
            with open(icon_file, "w", encoding="utf-8") as f:
                f.write(self.addon_to_install.macro.xpm)
            pixmapText = icon_file
        else:
            pixmapText = None

        # Add this command to that toolbar
        self._create_custom_command(
            toolbar,
            self.addon_to_install.macro.filename,
            menuText,
            tooltipText,
            whatsThisText,
            statustipText,
            pixmapText,
        )

    # pylint: disable=too-many-arguments
    def _create_custom_command(
        self,
        toolbar,
        filename,
        menuText,
        tooltipText,
        whatsThisText,
        statustipText,
        pixmapText,
    ):
        """Wrap createCustomCommand so it can be overridden during testing."""
        # Add this command to that toolbar
        command_name = FreeCADGui.Command.createCustomCommand(
            filename,
            menuText,
            tooltipText,
            whatsThisText,
            statustipText,
            pixmapText,
        )
        toolbar.SetString(command_name, "FreeCAD")

        # Force the toolbars to be recreated
        wb = FreeCADGui.activeWorkbench()
        wb.reloadActive()

    def _remove_custom_toolbar_button(self) -> None:
        """If this repo contains a macro, look through the custom commands and
        see if one is set up for this macro. If so, remove it, including any
        toolbar entries."""

        command = FreeCADGui.Command.findCustomCommand(self.addon_to_install.macro.filename)
        if not command:
            return
        custom_toolbars = FreeCAD.ParamGet("User parameter:BaseApp/Workbench/Global/Toolbar")
        toolbar_groups = custom_toolbars.GetGroups()
        for group in toolbar_groups:
            toolbar = custom_toolbars.GetGroup(group)
            if toolbar.GetString(command, "*") != "*":
                toolbar.RemString(command)

        FreeCADGui.Command.removeCustomCommand(command)

        # Force the toolbars to be recreated
        wb = FreeCADGui.activeWorkbench()
        wb.reloadActive()



# ***************************************************************************
# *   Copyright (c) 2014 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Lesser General Public License for more details.                   *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************


class PathCommandGroup:
    def __init__(self, cmdlist, menu, tooltip=None):
        self.cmdlist = cmdlist
        self.menu = menu
        if tooltip is None:
            self.tooltip = menu
        else:
            self.tooltip = tooltip

    def GetCommands(self):
        return tuple(self.cmdlist)

    def GetResources(self):
        return {"MenuText": self.menu, "ToolTip": self.tooltip}

    def IsActive(self):
        if FreeCAD.ActiveDocument is not None:
            for o in FreeCAD.ActiveDocument.Objects:
                if o.Name[:3] == "Job":
                    return True
        return False


class PathWorkbench(Workbench):
    "Path workbench"

    def __init__(self):
        self.__class__.Icon = (
            FreeCAD.getResourceDir() + "Mod/Path/Resources/icons/PathWorkbench.svg"
        )
        self.__class__.MenuText = "Path"
        self.__class__.ToolTip = "Path workbench"

    def Initialize(self):
        global PathCommandGroup

        # Add preferences pages - before loading PathGui to properly order pages of Path group
        import Path.Dressup.Gui.Preferences as PathPreferencesPathDressup
        import Path.Main.Gui.PreferencesJob as PathPreferencesPathJob

        translate = FreeCAD.Qt.translate

        # load the builtin modules
        import Path
        import PathScripts
        import PathGui
        from PySide import QtCore, QtGui

        FreeCADGui.addLanguagePath(":/translations")
        FreeCADGui.addIconPath(":/icons")
        import Path.GuiInit

        from Path.Main.Gui import JobCmd as PathJobCmd
        from Path.Tool.Gui import BitCmd as PathToolBitCmd
        from Path.Tool.Gui import BitLibraryCmd as PathToolBitLibraryCmd

        from PySide.QtCore import QT_TRANSLATE_NOOP

        import PathCommands
        import subprocess
        from packaging.version import Version, parse

        FreeCADGui.addPreferencePage(PathPreferencesPathJob.JobPreferencesPage, QT_TRANSLATE_NOOP("QObject", "Path"))
        FreeCADGui.addPreferencePage(
            PathPreferencesPathDressup.DressupPreferencesPage, QT_TRANSLATE_NOOP("QObject", "Path")
        )

        Path.GuiInit.Startup()

        # build commands list
        projcmdlist = ["Path_Job", "Path_Post", "Path_Sanity"]
        toolcmdlist = [
            "Path_Inspect",
            "Path_Simulator",
            "Path_SelectLoop",
            "Path_OpActiveToggle",
        ]
        prepcmdlist = [
            "Path_Fixture",
            "Path_Comment",
            "Path_Stop",
            "Path_Custom",
            "Path_Probe",
        ]
        twodopcmdlist = [
            "Path_Profile",
            "Path_Pocket_Shape",
            "Path_Drilling",
            "Path_MillFace",
            "Path_Helix",
            "Path_Adaptive",
        ]
        threedopcmdlist = ["Path_Pocket3D"]
        engravecmdlist = ["Path_Engrave", "Path_Deburr", "Path_Vcarve"]
        modcmdlist = ["Path_OperationCopy", "Path_Array", "Path_SimpleCopy"]
        dressupcmdlist = [
            "Path_DressupAxisMap",
            "Path_DressupPathBoundary",
            "Path_DressupDogbone",
            "Path_DressupDragKnife",
            "Path_DressupLeadInOut",
            "Path_DressupRampEntry",
            "Path_DressupTag",
            "Path_DressupZCorrect",
        ]
        extracmdlist = []
        # modcmdmore = ["Path_Hop",]
        # remotecmdlist = ["Path_Remote"]
        specialcmdlist = []

        toolcmdlist.extend(PathToolBitLibraryCmd.BarList)
        toolbitcmdlist = PathToolBitLibraryCmd.MenuList

        engravecmdgroup = ["Path_EngraveTools"]
        FreeCADGui.addCommand(
            "Path_EngraveTools",
            PathCommandGroup(
                engravecmdlist,
                QT_TRANSLATE_NOOP("Path_EngraveTools", "Engraving Operations"),
            ),
        )

        threedcmdgroup = threedopcmdlist
        if Path.Preferences.experimentalFeaturesEnabled():
            prepcmdlist.append("Path_Shape")
            extracmdlist.extend(["Path_Area", "Path_Area_Workplane"])
            specialcmdlist.append("Path_ThreadMilling")
            twodopcmdlist.append("Path_Slot")

        if Path.Preferences.advancedOCLFeaturesEnabled():
            try:
                r = subprocess.run(
                    ["camotics", "--version"], capture_output=True, text=True
                ).stderr.strip()
                v = parse(r)

                if v >= Version("1.2.2"):
                    toolcmdlist.append("Path_Camotics")
            except (FileNotFoundError, ModuleNotFoundError):
                pass

            try:
                try:
                    import ocl  # pylint: disable=unused-variable
                except ImportError:
                    import opencamlib as ocl
                from Path.Op.Gui import Surface
                from Path.Op.Gui import Waterline

                threedopcmdlist.extend(["Path_Surface", "Path_Waterline"])
                threedcmdgroup = ["Path_3dTools"]
                FreeCADGui.addCommand(
                    "Path_3dTools",
                    PathCommandGroup(
                        threedopcmdlist,
                        QT_TRANSLATE_NOOP("Path_3dTools", "3D Operations"),
                    ),
                )
            except ImportError:
                if not Path.Preferences.suppressOpenCamLibWarning():
                    FreeCAD.Console.PrintError("OpenCamLib is not working!\n")

        self.appendToolbar(QT_TRANSLATE_NOOP("Workbench", "Project Setup"), projcmdlist)
        self.appendToolbar(QT_TRANSLATE_NOOP("Workbench", "Tool Commands"), toolcmdlist)
        self.appendToolbar(
            QT_TRANSLATE_NOOP("Workbench", "New Operations"),
            twodopcmdlist + engravecmdgroup + threedcmdgroup,
        )
        self.appendToolbar(
            QT_TRANSLATE_NOOP("Workbench", "Path Modification"), modcmdlist
        )
        if extracmdlist:
            self.appendToolbar(
                QT_TRANSLATE_NOOP("Workbench", "Helpful Tools"), extracmdlist
            )

        self.appendMenu(
            [QT_TRANSLATE_NOOP("Workbench", "&Path")],
            projcmdlist
            + ["Path_ExportTemplate", "Separator"]
            + toolcmdlist
            + toolbitcmdlist
            + ["Separator"]
            + twodopcmdlist
            + engravecmdlist
            + ["Separator"]
            + threedopcmdlist
            + ["Separator"],
        )
        self.appendMenu(
            [
                QT_TRANSLATE_NOOP("Workbench", "&Path"),
                QT_TRANSLATE_NOOP("Workbench", "Path Dressup"),
            ],
            dressupcmdlist,
        )
        self.appendMenu(
            [
                QT_TRANSLATE_NOOP("Workbench", "&Path"),
                QT_TRANSLATE_NOOP("Workbench", "Supplemental Commands"),
            ],
            prepcmdlist,
        )
        self.appendMenu(
            [
                QT_TRANSLATE_NOOP("Workbench", "&Path"),
                QT_TRANSLATE_NOOP("Workbench", "Path Modification"),
            ],
            modcmdlist,
        )
        if specialcmdlist:
            self.appendMenu(
                [
                    QT_TRANSLATE_NOOP("Workbench", "&Path"),
                    QT_TRANSLATE_NOOP("Workbench", "Specialty Operations"),
                ],
                specialcmdlist,
            )
        if extracmdlist:
            self.appendMenu([QT_TRANSLATE_NOOP("Workbench", "&Path")], extracmdlist)

        self.appendMenu([QT_TRANSLATE_NOOP("Workbench", "&Path")], ["Separator"])
        self.appendMenu(
            [
                QT_TRANSLATE_NOOP("Workbench", "&Path"),
                QT_TRANSLATE_NOOP("Workbench", "Utils"),
            ],
            ["Path_PropertyBag"],
        )

        self.dressupcmds = dressupcmdlist

        curveAccuracy = Path.Preferences.defaultLibAreaCurveAccuracy()
        if curveAccuracy:
            Path.Area.setDefaultParams(Accuracy=curveAccuracy)

        # keep this one the last entry in the preferences
        import Path.Base.Gui.PreferencesAdvanced as PathPreferencesAdvanced
        from Path.Preferences import preferences

        FreeCADGui.addPreferencePage(
            PathPreferencesAdvanced.AdvancedPreferencesPage, QT_TRANSLATE_NOOP("QObject", "Path")
        )
        Log("Loading Path workbench... done\n")

        # Warn user if current schema doesn't use minute for time in velocity
        if not Path.Preferences.suppressVelocity():
            velString = FreeCAD.Units.Quantity(
                1, FreeCAD.Units.Velocity
            ).getUserPreferred()[2][3:]

            if velString != "min":
                current_schema = FreeCAD.Units.listSchemas(FreeCAD.Units.getSchema())

                msg = translate(
                    "Path",
                    "The currently selected unit schema: \n     '{}'\n Does not use 'minutes' for velocity values. \n \nCNC machines require feed rate to be expressed in \nunit/minute. To ensure correct G-code: \nSelect a minute-based schema in preferences.\nFor example:\n    'Metric, Small Parts & CNC'\n    'US Customary'\n    'Imperial Decimal'",
                ).format(current_schema)
                header = translate("Path", "Warning")
                msgbox = QtGui.QMessageBox(QtGui.QMessageBox.Warning, header, msg)

                msgbox.addButton(translate("Path", "Ok"), QtGui.QMessageBox.AcceptRole)
                msgbox.addButton(
                    translate("Path", "Don't Show This Anymore"),
                    QtGui.QMessageBox.ActionRole,
                )
                if msgbox.exec_() == 1:
                    preferences().SetBool("WarningSuppressVelocity", True)

    def GetClassName(self):
        return "Gui::PythonWorkbench"

    def Activated(self):
        # update the translation engine
        FreeCADGui.updateLocale()
        # Msg("Path workbench activated\n")

    def Deactivated(self):
        # Msg("Path workbench deactivated\n")
        pass

    def ContextMenu(self, recipient):
        import PathScripts

        menuAppended = False
        if len(FreeCADGui.Selection.getSelection()) == 1:
            obj = FreeCADGui.Selection.getSelection()[0]
            if obj.isDerivedFrom("Path::Feature"):
                self.appendContextMenu("", "Separator")
                self.appendContextMenu("", ["Path_Inspect"])
                selectedName = obj.Name
                if "Remote" in selectedName:
                    self.appendContextMenu("", ["Refresh_Path"])
                if "Job" in selectedName:
                    self.appendContextMenu(
                        "", ["Path_ExportTemplate"] + self.toolbitctxmenu
                    )
                menuAppended = True
            if isinstance(obj.Proxy, Path.Op.Base.ObjectOp):
                self.appendContextMenu(
                    "", ["Path_OperationCopy", "Path_OpActiveToggle"]
                )
                menuAppended = True
            if obj.isDerivedFrom("Path::Feature"):
                if (
                    "Profile" in selectedName
                    or "Contour" in selectedName
                    or "Dressup" in selectedName
                ):
                    self.appendContextMenu("", "Separator")
                    # self.appendContextMenu("", ["Set_StartPoint"])
                    # self.appendContextMenu("", ["Set_EndPoint"])
                    for cmd in self.dressupcmds:
                        self.appendContextMenu("", [cmd])
                    menuAppended = True
            if isinstance(obj.Proxy, Path.Tool.Bit.ToolBit):
                self.appendContextMenu("", ["Path_ToolBitSave", "Path_ToolBitSaveAs"])
                menuAppended = True
        if menuAppended:
            self.appendContextMenu("", "Separator")


Gui.addWorkbench(PathWorkbench())

FreeCAD.addImportType("GCode (*.nc *.NC *.gc *.GC *.ncc *.NCC *.ngc *.NGC *.cnc *.CNC *.tap *.TAP *.gcode *.GCODE)", "PathGui")



# -*- coding: utf-8 -*-
# ***************************************************************************
# *   Copyright (c) 2018 Kresimir Tusek <kresimir.tusek@gmail.com>          *
# *   Copyright (c) 2019-2021 Schildkroet                                   *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This library is free software; you can redistribute it and/or         *
# *   modify it under the terms of the GNU Library General Public           *
# *   License as published by the Free Software Foundation; either          *
# *   version 2 of the License, or (at your option) any later version.      *
# *                                                                         *
# *   This library  is distributed in the hope that it will be useful,      *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this library; see the file COPYING.LIB. If not,    *
# *   write to the Free Software Foundation, Inc., 59 Temple Place,         *
# *   Suite 330, Boston, MA  02111-1307, USA                                *
# *                                                                         *
# ***************************************************************************

import Path
import Path.Op.Base as PathOp
import PathScripts.PathUtils as PathUtils
import FreeCAD
import time
import json
import math
import area
from PySide.QtCore import QT_TRANSLATE_NOOP

if FreeCAD.GuiUp:
    from pivy import coin
    import FreeCADGui

__doc__ = "Class and implementation of the Adaptive path operation."

# lazily loaded modules
from lazy_loader.lazy_loader import LazyLoader

Part = LazyLoader("Part", globals(), "Part")
# TechDraw = LazyLoader('TechDraw', globals(), 'TechDraw')
FeatureExtensions = LazyLoader(
    "Path.Op.FeatureExtension", globals(), "Path.Op.FeatureExtension"
)
DraftGeomUtils = LazyLoader("DraftGeomUtils", globals(), "DraftGeomUtils")


if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


translate = FreeCAD.Qt.translate


def convertTo2d(pathArray):
    output = []
    for path in pathArray:
        pth2 = []
        for edge in path:
            for pt in edge:
                pth2.append([pt[0], pt[1]])
        output.append(pth2)
    return output


sceneGraph = None
scenePathNodes = []  # for scene cleanup aftewards
topZ = 10


def sceneDrawPath(path, color=(0, 0, 1)):
    coPoint = coin.SoCoordinate3()

    pts = []
    for pt in path:
        pts.append([pt[0], pt[1], topZ])

    coPoint.point.setValues(0, len(pts), pts)
    ma = coin.SoBaseColor()
    ma.rgb = color
    li = coin.SoLineSet()
    li.numVertices.setValue(len(pts))
    pathNode = coin.SoSeparator()
    pathNode.addChild(coPoint)
    pathNode.addChild(ma)
    pathNode.addChild(li)
    sceneGraph.addChild(pathNode)
    scenePathNodes.append(pathNode)  # for scene cleanup afterwards


def sceneClean():
    for n in scenePathNodes:
        sceneGraph.removeChild(n)

    del scenePathNodes[:]


def discretize(edge, flipDirection=False):
    pts = edge.discretize(Deflection=0.002)
    if flipDirection:
        pts.reverse()

    return pts


def CalcHelixConePoint(height, cur_z, radius, angle):
    x = ((height - cur_z) / height) * radius * math.cos(math.radians(angle) * cur_z)
    y = ((height - cur_z) / height) * radius * math.sin(math.radians(angle) * cur_z)
    z = cur_z

    return {"X": x, "Y": y, "Z": z}


def GenerateGCode(op, obj, adaptiveResults, helixDiameter):
    if len(adaptiveResults) == 0 or len(adaptiveResults[0]["AdaptivePaths"]) == 0:
        return

    # minLiftDistance = op.tool.Diameter
    helixRadius = 0
    for region in adaptiveResults:
        p1 = region["HelixCenterPoint"]
        p2 = region["StartPoint"]
        r = math.sqrt(
            (p1[0] - p2[0]) * (p1[0] - p2[0]) + (p1[1] - p2[1]) * (p1[1] - p2[1])
        )
        if r > helixRadius:
            helixRadius = r

    stepDown = obj.StepDown.Value
    passStartDepth = obj.StartDepth.Value

    if stepDown < 0.1:
        stepDown = 0.1

    length = 2 * math.pi * helixRadius

    if float(obj.HelixAngle) < 1:
        obj.HelixAngle = 1
    if float(obj.HelixAngle) > 89:
        obj.HelixAngle = 89

    if float(obj.HelixConeAngle) < 0:
        obj.HelixConeAngle = 0

    helixAngleRad = math.pi * float(obj.HelixAngle) / 180.0
    depthPerOneCircle = length * math.tan(helixAngleRad)
    # print("Helix circle depth: {}".format(depthPerOneCircle))

    stepUp = obj.LiftDistance.Value
    if stepUp < 0:
        stepUp = 0

    finish_step = obj.FinishDepth.Value if hasattr(obj, "FinishDepth") else 0.0
    if finish_step > stepDown:
        finish_step = stepDown

    depth_params = PathUtils.depth_params(
        clearance_height=obj.ClearanceHeight.Value,
        safe_height=obj.SafeHeight.Value,
        start_depth=obj.StartDepth.Value,
        step_down=stepDown,
        z_finish_step=finish_step,
        final_depth=obj.FinalDepth.Value,
        user_depths=None,
    )

    # ml: this is dangerous because it'll hide all unused variables hence forward
    #     however, I don't know what lx and ly signify so I'll leave them for now
    # lx = adaptiveResults[0]["HelixCenterPoint"][0]
    # ly = adaptiveResults[0]["HelixCenterPoint"][1]
    lz = passStartDepth
    step = 0

    for passEndDepth in depth_params.data:
        step = step + 1

        for region in adaptiveResults:
            startAngle = math.atan2(
                region["StartPoint"][1] - region["HelixCenterPoint"][1],
                region["StartPoint"][0] - region["HelixCenterPoint"][0],
            )

            # lx = region["HelixCenterPoint"][0]
            # ly = region["HelixCenterPoint"][1]

            passDepth = passStartDepth - passEndDepth

            p1 = region["HelixCenterPoint"]
            p2 = region["StartPoint"]
            helixRadius = math.sqrt(
                (p1[0] - p2[0]) * (p1[0] - p2[0]) + (p1[1] - p2[1]) * (p1[1] - p2[1])
            )

            # Helix ramp
            if helixRadius > 0.01:
                r = helixRadius - 0.01

                maxfi = passDepth / depthPerOneCircle * 2 * math.pi
                fi = 0
                offsetFi = -maxfi + startAngle - math.pi / 16

                helixStart = [
                    region["HelixCenterPoint"][0] + r * math.cos(offsetFi),
                    region["HelixCenterPoint"][1] + r * math.sin(offsetFi),
                ]

                op.commandlist.append(
                    Path.Command("(Helix to depth: %f)" % passEndDepth)
                )

                if obj.UseHelixArcs is False:
                    # rapid move to start point
                    op.commandlist.append(
                        Path.Command("G0", {"Z": obj.ClearanceHeight.Value})
                    )
                    op.commandlist.append(
                        Path.Command(
                            "G0",
                            {
                                "X": helixStart[0],
                                "Y": helixStart[1],
                                "Z": obj.ClearanceHeight.Value,
                            },
                        )
                    )

                    # rapid move to safe height
                    op.commandlist.append(
                        Path.Command(
                            "G0",
                            {
                                "X": helixStart[0],
                                "Y": helixStart[1],
                                "Z": obj.SafeHeight.Value,
                            },
                        )
                    )

                    # move to start depth
                    op.commandlist.append(
                        Path.Command(
                            "G1",
                            {
                                "X": helixStart[0],
                                "Y": helixStart[1],
                                "Z": passStartDepth,
                                "F": op.vertFeed,
                            },
                        )
                    )

                    if obj.HelixConeAngle == 0:
                        while fi < maxfi:
                            x = region["HelixCenterPoint"][0] + r * math.cos(
                                fi + offsetFi
                            )
                            y = region["HelixCenterPoint"][1] + r * math.sin(
                                fi + offsetFi
                            )
                            z = passStartDepth - fi / maxfi * (
                                passStartDepth - passEndDepth
                            )
                            op.commandlist.append(
                                Path.Command(
                                    "G1", {"X": x, "Y": y, "Z": z, "F": op.vertFeed}
                                )
                            )
                            # lx = x
                            # ly = y
                            fi = fi + math.pi / 16

                        # one more circle at target depth to make sure center is cleared
                        maxfi = maxfi + 2 * math.pi
                        while fi < maxfi:
                            x = region["HelixCenterPoint"][0] + r * math.cos(
                                fi + offsetFi
                            )
                            y = region["HelixCenterPoint"][1] + r * math.sin(
                                fi + offsetFi
                            )
                            z = passEndDepth
                            op.commandlist.append(
                                Path.Command(
                                    "G1", {"X": x, "Y": y, "Z": z, "F": op.horizFeed}
                                )
                            )
                            # lx = x
                            # ly = y
                            fi = fi + math.pi / 16

                    else:
                        # Cone
                        _HelixAngle = 360 - (float(obj.HelixAngle) * 4)

                        if obj.HelixConeAngle > 6:
                            obj.HelixConeAngle = 6

                        helixRadius *= 0.9

                        # Calculate everything
                        helix_height = passStartDepth - passEndDepth
                        r_extra = helix_height * math.tan(
                            math.radians(obj.HelixConeAngle)
                        )
                        HelixTopRadius = helixRadius + r_extra
                        helix_full_height = HelixTopRadius * (
                            math.cos(math.radians(obj.HelixConeAngle))
                            / math.sin(math.radians(obj.HelixConeAngle))
                        )

                        # Start height
                        z = passStartDepth
                        i = 0

                        # Default step down
                        z_step = 0.05

                        # Bigger angle, smaller step down
                        if _HelixAngle > 120:
                            z_step = 0.025
                        if _HelixAngle > 240:
                            z_step = 0.015

                        p = None
                        # Calculate conical helix
                        while z >= passEndDepth:
                            if z < passEndDepth:
                                z = passEndDepth

                            p = CalcHelixConePoint(
                                helix_full_height, i, HelixTopRadius, _HelixAngle
                            )
                            op.commandlist.append(
                                Path.Command(
                                    "G1",
                                    {
                                        "X": p["X"] + region["HelixCenterPoint"][0],
                                        "Y": p["Y"] + region["HelixCenterPoint"][1],
                                        "Z": z,
                                        "F": op.vertFeed,
                                    },
                                )
                            )
                            z = z - z_step
                            i = i + z_step

                        # Calculate some stuff for arcs at bottom
                        p["X"] = p["X"] + region["HelixCenterPoint"][0]
                        p["Y"] = p["Y"] + region["HelixCenterPoint"][1]
                        x_m = (
                            region["HelixCenterPoint"][0]
                            - p["X"]
                            + region["HelixCenterPoint"][0]
                        )
                        y_m = (
                            region["HelixCenterPoint"][1]
                            - p["Y"]
                            + region["HelixCenterPoint"][1]
                        )
                        i_off = (x_m - p["X"]) / 2
                        j_off = (y_m - p["Y"]) / 2

                        # One more circle at target depth to make sure center is cleared
                        op.commandlist.append(
                            Path.Command(
                                "G3",
                                {
                                    "X": x_m,
                                    "Y": y_m,
                                    "Z": passEndDepth,
                                    "I": i_off,
                                    "J": j_off,
                                    "F": op.horizFeed,
                                },
                            )
                        )
                        op.commandlist.append(
                            Path.Command(
                                "G3",
                                {
                                    "X": p["X"],
                                    "Y": p["Y"],
                                    "Z": passEndDepth,
                                    "I": -i_off,
                                    "J": -j_off,
                                    "F": op.horizFeed,
                                },
                            )
                        )

                else:
                    # Use arcs for helix - no conical shape support
                    helixStart = [
                        region["HelixCenterPoint"][0] + r,
                        region["HelixCenterPoint"][1],
                    ]

                    # rapid move to start point
                    op.commandlist.append(
                        Path.Command("G0", {"Z": obj.ClearanceHeight.Value})
                    )
                    op.commandlist.append(
                        Path.Command(
                            "G0",
                            {
                                "X": helixStart[0],
                                "Y": helixStart[1],
                                "Z": obj.ClearanceHeight.Value,
                            },
                        )
                    )

                    # rapid move to safe height
                    op.commandlist.append(
                        Path.Command(
                            "G0",
                            {
                                "X": helixStart[0],
                                "Y": helixStart[1],
                                "Z": obj.SafeHeight.Value,
                            },
                        )
                    )

                    # move to start depth
                    op.commandlist.append(
                        Path.Command(
                            "G1",
                            {
                                "X": helixStart[0],
                                "Y": helixStart[1],
                                "Z": passStartDepth,
                                "F": op.vertFeed,
                            },
                        )
                    )

                    x = region["HelixCenterPoint"][0] + r
                    y = region["HelixCenterPoint"][1]

                    curDep = passStartDepth
                    while curDep > (passEndDepth + depthPerOneCircle):
                        op.commandlist.append(
                            Path.Command(
                                "G2",
                                {
                                    "X": x - (2 * r),
                                    "Y": y,
                                    "Z": curDep - (depthPerOneCircle / 2),
                                    "I": -r,
                                    "F": op.vertFeed,
                                },
                            )
                        )
                        op.commandlist.append(
                            Path.Command(
                                "G2",
                                {
                                    "X": x,
                                    "Y": y,
                                    "Z": curDep - depthPerOneCircle,
                                    "I": r,
                                    "F": op.vertFeed,
                                },
                            )
                        )
                        curDep = curDep - depthPerOneCircle

                    lastStep = curDep - passEndDepth
                    if lastStep > (depthPerOneCircle / 2):
                        op.commandlist.append(
                            Path.Command(
                                "G2",
                                {
                                    "X": x - (2 * r),
                                    "Y": y,
                                    "Z": curDep - (lastStep / 2),
                                    "I": -r,
                                    "F": op.vertFeed,
                                },
                            )
                        )
                        op.commandlist.append(
                            Path.Command(
                                "G2",
                                {
                                    "X": x,
                                    "Y": y,
                                    "Z": passEndDepth,
                                    "I": r,
                                    "F": op.vertFeed,
                                },
                            )
                        )
                    else:
                        op.commandlist.append(
                            Path.Command(
                                "G2",
                                {
                                    "X": x - (2 * r),
                                    "Y": y,
                                    "Z": passEndDepth,
                                    "I": -r,
                                    "F": op.vertFeed,
                                },
                            )
                        )
                        op.commandlist.append(
                            Path.Command(
                                "G1",
                                {"X": x, "Y": y, "Z": passEndDepth, "F": op.vertFeed},
                            )
                        )

                    # one more circle at target depth to make sure center is cleared
                    op.commandlist.append(
                        Path.Command(
                            "G2",
                            {
                                "X": x - (2 * r),
                                "Y": y,
                                "Z": passEndDepth,
                                "I": -r,
                                "F": op.horizFeed,
                            },
                        )
                    )
                    op.commandlist.append(
                        Path.Command(
                            "G2",
                            {
                                "X": x,
                                "Y": y,
                                "Z": passEndDepth,
                                "I": r,
                                "F": op.horizFeed,
                            },
                        )
                    )
                    # lx = x
                    # ly = y

            else:  # no helix entry
                # rapid move to clearance height
                op.commandlist.append(
                    Path.Command("G0", {"Z": obj.ClearanceHeight.Value})
                )
                op.commandlist.append(
                    Path.Command(
                        "G0",
                        {
                            "X": region["StartPoint"][0],
                            "Y": region["StartPoint"][1],
                            "Z": obj.ClearanceHeight.Value,
                        },
                    )
                )
                # straight plunge to target depth
                op.commandlist.append(
                    Path.Command(
                        "G1",
                        {
                            "X": region["StartPoint"][0],
                            "Y": region["StartPoint"][1],
                            "Z": passEndDepth,
                            "F": op.vertFeed,
                        },
                    )
                )

            lz = passEndDepth
            z = obj.ClearanceHeight.Value
            op.commandlist.append(Path.Command("(Adaptive - depth: %f)" % passEndDepth))

            # add adaptive paths
            for pth in region["AdaptivePaths"]:
                motionType = pth[0]  # [0] contains motion type

                for pt in pth[1]:  # [1] contains list of points
                    x = pt[0]
                    y = pt[1]

                    # dist = math.sqrt((x-lx)*(x-lx) + (y-ly)*(y-ly))

                    if motionType == area.AdaptiveMotionType.Cutting:
                        z = passEndDepth
                        if z != lz:
                            op.commandlist.append(
                                Path.Command("G1", {"Z": z, "F": op.vertFeed})
                            )

                        op.commandlist.append(
                            Path.Command("G1", {"X": x, "Y": y, "F": op.horizFeed})
                        )

                    elif motionType == area.AdaptiveMotionType.LinkClear:
                        z = passEndDepth + stepUp
                        if z != lz:
                            op.commandlist.append(Path.Command("G0", {"Z": z}))

                        op.commandlist.append(Path.Command("G0", {"X": x, "Y": y}))

                    elif motionType == area.AdaptiveMotionType.LinkNotClear:
                        z = obj.ClearanceHeight.Value
                        if z != lz:
                            op.commandlist.append(Path.Command("G0", {"Z": z}))

                        op.commandlist.append(Path.Command("G0", {"X": x, "Y": y}))

                    # elif motionType == area.AdaptiveMotionType.LinkClearAtPrevPass:
                    #     if lx!=x or ly!=y:
                    #         op.commandlist.append(Path.Command("G0", { "X": lx, "Y":ly, "Z":passStartDepth+stepUp}))
                    #     op.commandlist.append(Path.Command("G0", { "X": x, "Y":y, "Z":passStartDepth+stepUp}))

                    # lx = x
                    # ly = y
                    lz = z

            # return to safe height in this Z pass
            z = obj.ClearanceHeight.Value
            if z != lz:
                op.commandlist.append(Path.Command("G0", {"Z": z}))

            lz = z

        passStartDepth = passEndDepth

        # return to safe height in this Z pass
        z = obj.ClearanceHeight.Value
        if z != lz:
            op.commandlist.append(Path.Command("G0", {"Z": z}))

        lz = z

    z = obj.ClearanceHeight.Value
    if z != lz:
        op.commandlist.append(Path.Command("G0", {"Z": z}))


def Execute(op, obj):
    global sceneGraph
    global topZ

    if FreeCAD.GuiUp:
        sceneGraph = FreeCADGui.ActiveDocument.ActiveView.getSceneGraph()

    Path.Log.info("*** Adaptive toolpath processing started...\n")

    # hide old toolpaths during recalculation
    obj.Path = Path.Path("(Calculating...)")

    if FreeCAD.GuiUp:
        # store old visibility state
        job = op.getJob(obj)
        oldObjVisibility = obj.ViewObject.Visibility
        oldJobVisibility = job.ViewObject.Visibility

        obj.ViewObject.Visibility = False
        job.ViewObject.Visibility = False

        FreeCADGui.updateGui()

    try:
        helixDiameter = obj.HelixDiameterLimit.Value
        topZ = op.stock.Shape.BoundBox.ZMax
        obj.Stopped = False
        obj.StopProcessing = False
        if obj.Tolerance < 0.001:
            obj.Tolerance = 0.001

        # Get list of working edges for adaptive algorithm
        pathArray = op.pathArray
        if not pathArray:
            Path.Log.error("No wire data returned.")
            return

        path2d = convertTo2d(pathArray)

        stockPaths = []
        if hasattr(op.stock, "StockType") and op.stock.StockType == "CreateCylinder":
            stockPaths.append([discretize(op.stock.Shape.Edges[0])])

        else:
            stockBB = op.stock.Shape.BoundBox
            v = []
            v.append(FreeCAD.Vector(stockBB.XMin, stockBB.YMin, 0))
            v.append(FreeCAD.Vector(stockBB.XMax, stockBB.YMin, 0))
            v.append(FreeCAD.Vector(stockBB.XMax, stockBB.YMax, 0))
            v.append(FreeCAD.Vector(stockBB.XMin, stockBB.YMax, 0))
            v.append(FreeCAD.Vector(stockBB.XMin, stockBB.YMin, 0))
            stockPaths.append([v])

        stockPath2d = convertTo2d(stockPaths)

        # opType = area.AdaptiveOperationType.ClearingInside  # Commented out per LGTM suggestion
        if obj.OperationType == "Clearing":
            if obj.Side == "Outside":
                opType = area.AdaptiveOperationType.ClearingOutside

            else:
                opType = area.AdaptiveOperationType.ClearingInside

        else:  # profiling
            if obj.Side == "Outside":
                opType = area.AdaptiveOperationType.ProfilingOutside

            else:
                opType = area.AdaptiveOperationType.ProfilingInside

        keepToolDownRatio = 3.0
        if hasattr(obj, "KeepToolDownRatio"):
            keepToolDownRatio = float(obj.KeepToolDownRatio)

        # put here all properties that influence calculation of adaptive base paths,

        inputStateObject = {
            "tool": float(op.tool.Diameter),
            "tolerance": float(obj.Tolerance),
            "geometry": path2d,
            "stockGeometry": stockPath2d,
            "stepover": float(obj.StepOver),
            "effectiveHelixDiameter": float(helixDiameter),
            "operationType": obj.OperationType,
            "side": obj.Side,
            "forceInsideOut": obj.ForceInsideOut,
            "finishingProfile": obj.FinishingProfile,
            "keepToolDownRatio": keepToolDownRatio,
            "stockToLeave": float(obj.StockToLeave),
        }

        inputStateChanged = False
        adaptiveResults = None

        if obj.AdaptiveOutputState is not None and obj.AdaptiveOutputState != "":
            adaptiveResults = obj.AdaptiveOutputState

        if json.dumps(obj.AdaptiveInputState) != json.dumps(inputStateObject):
            inputStateChanged = True
            adaptiveResults = None

        # progress callback fn, if return true it will stop processing
        def progressFn(tpaths):
            if FreeCAD.GuiUp:
                for (
                    path
                ) in (
                    tpaths
                ):  # path[0] contains the MotionType, #path[1] contains list of points
                    if path[0] == area.AdaptiveMotionType.Cutting:
                        sceneDrawPath(path[1], (0, 0, 1))

                    else:
                        sceneDrawPath(path[1], (1, 0, 1))

                FreeCADGui.updateGui()

            return obj.StopProcessing

        start = time.time()

        if inputStateChanged or adaptiveResults is None:
            a2d = area.Adaptive2d()
            a2d.stepOverFactor = 0.01 * obj.StepOver
            a2d.toolDiameter = float(op.tool.Diameter)
            a2d.helixRampDiameter = helixDiameter
            a2d.keepToolDownDistRatio = keepToolDownRatio
            a2d.stockToLeave = float(obj.StockToLeave)
            a2d.tolerance = float(obj.Tolerance)
            a2d.forceInsideOut = obj.ForceInsideOut
            a2d.finishingProfile = obj.FinishingProfile
            a2d.opType = opType

            # EXECUTE
            results = a2d.Execute(stockPath2d, path2d, progressFn)

            # need to convert results to python object to be JSON serializable
            adaptiveResults = []
            for result in results:
                adaptiveResults.append(
                    {
                        "HelixCenterPoint": result.HelixCenterPoint,
                        "StartPoint": result.StartPoint,
                        "AdaptivePaths": result.AdaptivePaths,
                        "ReturnMotionType": result.ReturnMotionType,
                    }
                )

        # GENERATE
        GenerateGCode(op, obj, adaptiveResults, helixDiameter)

        if not obj.StopProcessing:
            Path.Log.info("*** Done. Elapsed time: %f sec\n\n" % (time.time() - start))
            obj.AdaptiveOutputState = adaptiveResults
            obj.AdaptiveInputState = inputStateObject

        else:
            Path.Log.info(
                "*** Processing cancelled (after: %f sec).\n\n" % (time.time() - start)
            )

    finally:
        if FreeCAD.GuiUp:
            obj.ViewObject.Visibility = oldObjVisibility
            job.ViewObject.Visibility = oldJobVisibility
            sceneClean()


def _get_working_edges(op, obj):
    """_get_working_edges(op, obj)...
    Compile all working edges from the Base Geometry selection (obj.Base)
    for the current operation.
    Additional modifications to selected region(face), such as extensions,
    should be placed within this function.
    """
    all_regions = list()
    edge_list = list()
    avoidFeatures = list()
    rawEdges = list()

    # Get extensions and identify faces to avoid
    extensions = FeatureExtensions.getExtensions(obj)
    for e in extensions:
        if e.avoid:
            avoidFeatures.append(e.feature)

    # Get faces selected by user
    for base, subs in obj.Base:
        for sub in subs:
            if sub.startswith("Face"):
                if sub not in avoidFeatures:
                    if obj.UseOutline:
                        face = base.Shape.getElement(sub)
                        # get outline with wire_A method used in PocketShape, but it does not play nicely later
                        # wire_A = TechDraw.findShapeOutline(face, 1, FreeCAD.Vector(0.0, 0.0, 1.0))
                        wire_B = face.OuterWire
                        shape = Part.Face(wire_B)
                    else:
                        shape = base.Shape.getElement(sub)
                    all_regions.append(shape)
            elif sub.startswith("Edge"):
                # Save edges for later processing
                rawEdges.append(base.Shape.getElement(sub))
    # Efor

    # Process selected edges
    if rawEdges:
        edgeWires = DraftGeomUtils.findWires(rawEdges)
        if edgeWires:
            for w in edgeWires:
                for e in w.Edges:
                    edge_list.append([discretize(e)])

    # Apply regular Extensions
    op.exts = []
    for ext in extensions:
        if not ext.avoid:
            wire = ext.getWire()
            if wire:
                for f in ext.getExtensionFaces(wire):
                    op.exts.append(f)
                    all_regions.append(f)

    # Second face-combining method attempted
    horizontal = Path.Geom.combineHorizontalFaces(all_regions)
    if horizontal:
        obj.removalshape = Part.makeCompound(horizontal)
        for f in horizontal:
            for w in f.Wires:
                for e in w.Edges:
                    edge_list.append([discretize(e)])

    return edge_list


class PathAdaptive(PathOp.ObjectOp):
    def opFeatures(self, obj):
        """opFeatures(obj) ... returns the OR'ed list of features used and supported by the operation.
        The default implementation returns "FeatureTool | FeatureDepths | FeatureHeights | FeatureStartPoint"
        Should be overwritten by subclasses."""
        return (
            PathOp.FeatureTool
            | PathOp.FeatureBaseEdges
            | PathOp.FeatureDepths
            | PathOp.FeatureFinishDepth
            | PathOp.FeatureStepDown
            | PathOp.FeatureHeights
            | PathOp.FeatureBaseGeometry
            | PathOp.FeatureCoolant
            | PathOp.FeatureLocations
        )

    @classmethod
    def propertyEnumerations(self, dataType="data"):
        """helixOpPropertyEnumerations(dataType="data")... return property enumeration lists of specified dataType.
        Args:
            dataType = 'data', 'raw', 'translated'
        Notes:
        'data' is list of internal string literals used in code
        'raw' is list of (translated_text, data_string) tuples
        'translated' is list of translated string literals
        """

        # Enumeration lists for App::PropertyEnumeration properties
        enums = {
            "Side": [
                (translate("Path_Adaptive", "Outside"), "Outside"),
                (translate("Path_Adaptive", "Inside"), "Inside"),
            ],  # this is the direction that the profile runs
            "OperationType": [
                (translate("Path_Adaptive", "Clearing"), "Clearing"),
                (translate("Path_Adaptive", "Profiling"), "Profiling"),
            ],  # side of profile that cutter is on in relation to direction of profile
        }

        if dataType == "raw":
            return enums

        data = list()
        idx = 0 if dataType == "translated" else 1

        Path.Log.debug(enums)

        for k, v in enumerate(enums):
            data.append((v, [tup[idx] for tup in enums[v]]))
        Path.Log.debug(data)

        return data

    def initOperation(self, obj):
        """initOperation(obj) ... implement to create additional properties.
        Should be overwritten by subclasses."""
        obj.addProperty(
            "App::PropertyEnumeration",
            "Side",
            "Adaptive",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Side of selected faces that tool should cut",
            ),
        )
        # obj.Side = [
        #     "Outside",
        #     "Inside",
        # ]  # side of profile that cutter is on in relation to direction of profile

        obj.addProperty(
            "App::PropertyEnumeration",
            "OperationType",
            "Adaptive",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Type of adaptive operation",
            ),
        )
        # obj.OperationType = [
        #     "Clearing",
        #     "Profiling",
        # ]  # side of profile that cutter is on in relation to direction of profile

        obj.addProperty(
            "App::PropertyFloat",
            "Tolerance",
            "Adaptive",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Influences accuracy and performance",
            ),
        )
        obj.addProperty(
            "App::PropertyPercent",
            "StepOver",
            "Adaptive",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Percent of cutter diameter to step over on each pass",
            ),
        )
        obj.addProperty(
            "App::PropertyDistance",
            "LiftDistance",
            "Adaptive",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Lift distance for rapid moves",
            ),
        )
        obj.addProperty(
            "App::PropertyDistance",
            "KeepToolDownRatio",
            "Adaptive",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Max length of keep tool down path compared to direct distance between points",
            ),
        )
        obj.addProperty(
            "App::PropertyDistance",
            "StockToLeave",
            "Adaptive",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "How much stock to leave (i.e. for finishing operation)",
            ),
        )
        obj.addProperty(
            "App::PropertyBool",
            "ForceInsideOut",
            "Adaptive",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Force plunging into material inside and clearing towards the edges",
            ),
        )
        obj.addProperty(
            "App::PropertyBool",
            "FinishingProfile",
            "Adaptive",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "To take a finishing profile path at the end",
            ),
        )
        obj.addProperty(
            "App::PropertyBool",
            "Stopped",
            "Adaptive",
            QT_TRANSLATE_NOOP("App::Property", "Stop processing"),
        )
        obj.setEditorMode("Stopped", 2)  # hide this property

        obj.addProperty(
            "App::PropertyBool",
            "StopProcessing",
            "Adaptive",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Stop processing",
            ),
        )
        obj.setEditorMode("StopProcessing", 2)  # hide this property

        obj.addProperty(
            "App::PropertyBool",
            "UseHelixArcs",
            "Adaptive",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Use Arcs (G2) for helix ramp",
            ),
        )

        obj.addProperty(
            "App::PropertyPythonObject",
            "AdaptiveInputState",
            "Adaptive",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Internal input state",
            ),
        )
        obj.addProperty(
            "App::PropertyPythonObject",
            "AdaptiveOutputState",
            "Adaptive",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Internal output state",
            ),
        )
        obj.setEditorMode("AdaptiveInputState", 2)  # hide this property
        obj.setEditorMode("AdaptiveOutputState", 2)  # hide this property
        obj.addProperty(
            "App::PropertyAngle",
            "HelixAngle",
            "Adaptive",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Helix ramp entry angle (degrees)",
            ),
        )
        obj.addProperty(
            "App::PropertyAngle",
            "HelixConeAngle",
            "Adaptive",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Helix cone angle (degrees)",
            ),
        )
        obj.addProperty(
            "App::PropertyLength",
            "HelixDiameterLimit",
            "Adaptive",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Limit helix entry diameter, if limit larger than tool diameter or 0, tool diameter is used",
            ),
        )

        obj.addProperty(
            "App::PropertyBool",
            "UseOutline",
            "Adaptive",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Uses the outline of the base geometry.",
            ),
        )

        obj.addProperty(
            "Part::PropertyPartShape",
            "removalshape",
            "Path",
            QT_TRANSLATE_NOOP("App::Property", ""),
        )

        for n in self.propertyEnumerations():
            setattr(obj, n[0], n[1])

        obj.setEditorMode("removalshape", 2)  # hide

        FeatureExtensions.initialize_properties(obj)

    def opSetDefaultValues(self, obj, job):
        obj.Side = "Inside"
        obj.OperationType = "Clearing"
        obj.Tolerance = 0.1
        obj.StepOver = 20
        obj.LiftDistance = 0
        # obj.ProcessHoles = True
        obj.ForceInsideOut = False
        obj.FinishingProfile = True
        obj.Stopped = False
        obj.StopProcessing = False
        obj.HelixAngle = 5
        obj.HelixConeAngle = 0
        obj.HelixDiameterLimit = 0.0
        obj.AdaptiveInputState = ""
        obj.AdaptiveOutputState = ""
        obj.StockToLeave = 0
        obj.KeepToolDownRatio = 3.0
        obj.UseHelixArcs = False
        obj.UseOutline = False
        FeatureExtensions.set_default_property_values(obj, job)

    def opExecute(self, obj):
        """opExecute(obj) ... called whenever the receiver needs to be recalculated.
        See documentation of execute() for a list of base functionality provided.
        Should be overwritten by subclasses."""

        self.pathArray = _get_working_edges(self, obj)
        Execute(self, obj)

    def opOnDocumentRestored(self, obj):
        if not hasattr(obj, "HelixConeAngle"):
            obj.addProperty(
                "App::PropertyAngle",
                "HelixConeAngle",
                "Adaptive",
                "Helix cone angle (degrees)",
            )

        if not hasattr(obj, "UseOutline"):
            obj.addProperty(
                "App::PropertyBool",
                "UseOutline",
                "Adaptive",
                "Uses the outline of the base geometry.",
            )

        if not hasattr(obj, "removalshape"):
            obj.addProperty("Part::PropertyPartShape", "removalshape", "Path", "")
        obj.setEditorMode("removalshape", 2)  # hide

        FeatureExtensions.initialize_properties(obj)


# Eclass


def SetupProperties():
    setup = [
        "Side",
        "OperationType",
        "Tolerance",
        "StepOver",
        "LiftDistance",
        "KeepToolDownRatio",
        "StockToLeave",
        "ForceInsideOut",
        "FinishingProfile",
        "Stopped",
        "StopProcessing",
        "UseHelixArcs",
        "AdaptiveInputState",
        "AdaptiveOutputState",
        "HelixAngle",
        "HelixConeAngle",
        "HelixDiameterLimit",
        "UseOutline",
    ]
    return setup


def Create(name, obj=None, parentJob=None):
    """Create(name) ... Creates and returns a Adaptive operation."""
    if obj is None:
        obj = FreeCAD.ActiveDocument.addObject("Path::FeaturePython", name)
    obj.Proxy = PathAdaptive(obj, name, parentJob)
    return obj



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""Contains the unit test class for addonmanager_uninstaller.py non-GUI functionality."""

import functools
import os
from stat import S_IREAD, S_IRGRP, S_IROTH, S_IWUSR
import tempfile
import unittest

import FreeCAD

from addonmanager_uninstaller import AddonUninstaller, MacroUninstaller

from Addon import Addon
from AddonManagerTest.app.mocks import MockAddon, MockMacro


class TestAddonUninstaller(unittest.TestCase):
    """Test class for addonmanager_uninstaller.py non-GUI functionality"""

    MODULE = "test_uninstaller"  # file name without extension

    def setUp(self):
        """Initialize data needed for all tests"""
        self.test_data_dir = os.path.join(
            FreeCAD.getHomePath(), "Mod", "AddonManager", "AddonManagerTest", "data"
        )
        self.mock_addon = MockAddon()
        self.signals_caught = []
        self.test_object = AddonUninstaller(self.mock_addon)

        self.test_object.finished.connect(
            functools.partial(self.catch_signal, "finished")
        )
        self.test_object.success.connect(
            functools.partial(self.catch_signal, "success")
        )
        self.test_object.failure.connect(
            functools.partial(self.catch_signal, "failure")
        )

    def tearDown(self):
        """Finalize the test."""

    def catch_signal(self, signal_name, *_):
        """Internal use: used to catch and log any emitted signals. Not called directly."""
        self.signals_caught.append(signal_name)

    def setup_dummy_installation(self, temp_dir) -> str:
        """Set up a dummy Addon in temp_dir"""
        toplevel_path = os.path.join(temp_dir, self.mock_addon.name)
        os.makedirs(toplevel_path)
        with open(os.path.join(toplevel_path, "README.md"), "w") as f:
            f.write("## Mock Addon ##\n\nFile created by the unit test code.")
        self.test_object.installation_path = temp_dir
        return toplevel_path

    def create_fake_macro(self, macro_directory, fake_macro_name, digest):
        """Create an FCMacro file and matching digest entry for later removal"""
        os.makedirs(macro_directory, exist_ok=True)
        fake_file_installed = os.path.join(macro_directory, fake_macro_name)
        with open(digest, "a", encoding="utf-8") as f:
            f.write("# The following files were created outside this installation:\n")
            f.write(fake_file_installed + "\n")
        with open(fake_file_installed, "w", encoding="utf-8") as f:
            f.write("# Fake macro data for unit testing")

    def test_uninstall_normal(self):
        """Test the integrated uninstall function under normal circumstances"""

        with tempfile.TemporaryDirectory() as temp_dir:
            toplevel_path = self.setup_dummy_installation(temp_dir)
            self.test_object.run()
            self.assertTrue(os.path.exists(temp_dir))
            self.assertFalse(os.path.exists(toplevel_path))
            self.assertNotIn("failure", self.signals_caught)
            self.assertIn("success", self.signals_caught)
            self.assertIn("finished", self.signals_caught)

    def test_uninstall_no_name(self):
        """Test the integrated uninstall function for an addon without a name"""

        with tempfile.TemporaryDirectory() as temp_dir:
            toplevel_path = self.setup_dummy_installation(temp_dir)
            self.mock_addon.name = None
            result = self.test_object.run()
            self.assertTrue(os.path.exists(temp_dir))
            self.assertIn("failure", self.signals_caught)
            self.assertNotIn("success", self.signals_caught)
            self.assertIn("finished", self.signals_caught)

    def test_uninstall_dangerous_name(self):
        """Test the integrated uninstall function for an addon with a dangerous name"""

        with tempfile.TemporaryDirectory() as temp_dir:
            toplevel_path = self.setup_dummy_installation(temp_dir)
            self.mock_addon.name = "./"
            result = self.test_object.run()
            self.assertTrue(os.path.exists(temp_dir))
            self.assertIn("failure", self.signals_caught)
            self.assertNotIn("success", self.signals_caught)
            self.assertIn("finished", self.signals_caught)

    def test_uninstall_unmatching_name(self):
        """Test the integrated uninstall function for an addon with a name that isn't installed"""

        with tempfile.TemporaryDirectory() as temp_dir:
            toplevel_path = self.setup_dummy_installation(temp_dir)
            self.mock_addon.name += "Nonexistent"
            result = self.test_object.run()
            self.assertTrue(os.path.exists(temp_dir))
            self.assertIn("failure", self.signals_caught)
            self.assertNotIn("success", self.signals_caught)
            self.assertIn("finished", self.signals_caught)

    def test_uninstall_addon_with_macros(self):
        """Tests that the uninstaller removes the macro files"""
        with tempfile.TemporaryDirectory() as temp_dir:
            toplevel_path = self.setup_dummy_installation(temp_dir)
            macro_directory = os.path.join(temp_dir, "Macros")
            self.create_fake_macro(
                macro_directory,
                "FakeMacro.FCMacro",
                os.path.join(toplevel_path, "AM_INSTALLATION_DIGEST.txt"),
            )
            result = self.test_object.run()
            self.assertNotIn("failure", self.signals_caught)
            self.assertIn("success", self.signals_caught)
            self.assertIn("finished", self.signals_caught)
            self.assertFalse(
                os.path.exists(os.path.join(macro_directory, "FakeMacro.FCMacro"))
            )
            self.assertTrue(os.path.exists(macro_directory))

    def test_uninstall_calls_script(self):
        """Tests that the main uninstaller run function calls the uninstall.py script"""

        class Interceptor:
            def __init__(self):
                self.called = False
                self.args = []

            def func(self, *args):
                self.called = True
                self.args = args

        interceptor = Interceptor()
        with tempfile.TemporaryDirectory() as temp_dir:
            toplevel_path = self.setup_dummy_installation(temp_dir)
            self.test_object.run_uninstall_script = interceptor.func
            result = self.test_object.run()
            self.assertTrue(interceptor.called, "Failed to call uninstall script")

    def test_remove_extra_files_no_digest(self):
        """Tests that a lack of digest file is not an error, and nothing gets removed"""
        with tempfile.TemporaryDirectory() as temp_dir:
            self.test_object.remove_extra_files(temp_dir)  # Shouldn't throw
            self.assertTrue(os.path.exists(temp_dir))

    def test_remove_extra_files_empty_digest(self):
        """Test that an empty digest file is not an error, and nothing gets removed"""
        with tempfile.TemporaryDirectory() as temp_dir:
            with open("AM_INSTALLATION_DIGEST.txt", "w", encoding="utf-8") as f:
                f.write("")
            self.test_object.remove_extra_files(temp_dir)  # Shouldn't throw
            self.assertTrue(os.path.exists(temp_dir))

    def test_remove_extra_files_comment_only_digest(self):
        """Test that a digest file that contains only comment lines is not an error, and nothing
        gets removed"""
        with tempfile.TemporaryDirectory() as temp_dir:
            with open("AM_INSTALLATION_DIGEST.txt", "w", encoding="utf-8") as f:
                f.write("# Fake digest file for unit testing")
            self.test_object.remove_extra_files(temp_dir)  # Shouldn't throw
            self.assertTrue(os.path.exists(temp_dir))

    def test_remove_extra_files_repeated_files(self):
        """Test that a digest with the same file repeated removes it once, but doesn't error on
        later requests to remove it."""
        with tempfile.TemporaryDirectory() as temp_dir:
            toplevel_path = self.setup_dummy_installation(temp_dir)
            macro_directory = os.path.join(temp_dir, "Macros")
            self.create_fake_macro(
                macro_directory,
                "FakeMacro.FCMacro",
                os.path.join(toplevel_path, "AM_INSTALLATION_DIGEST.txt"),
            )
            self.create_fake_macro(
                macro_directory,
                "FakeMacro.FCMacro",
                os.path.join(toplevel_path, "AM_INSTALLATION_DIGEST.txt"),
            )
            self.create_fake_macro(
                macro_directory,
                "FakeMacro.FCMacro",
                os.path.join(toplevel_path, "AM_INSTALLATION_DIGEST.txt"),
            )
            self.test_object.remove_extra_files(toplevel_path)  # Shouldn't throw
            self.assertFalse(
                os.path.exists(os.path.join(macro_directory, "FakeMacro.FCMacro"))
            )

    def test_remove_extra_files_normal_case(self):
        """Test that a digest that is a "normal" case removes the requested files"""
        with tempfile.TemporaryDirectory() as temp_dir:
            toplevel_path = self.setup_dummy_installation(temp_dir)
            macro_directory = os.path.join(temp_dir, "Macros")
            self.create_fake_macro(
                macro_directory,
                "FakeMacro1.FCMacro",
                os.path.join(toplevel_path, "AM_INSTALLATION_DIGEST.txt"),
            )
            self.create_fake_macro(
                macro_directory,
                "FakeMacro2.FCMacro",
                os.path.join(toplevel_path, "AM_INSTALLATION_DIGEST.txt"),
            )
            self.create_fake_macro(
                macro_directory,
                "FakeMacro3.FCMacro",
                os.path.join(toplevel_path, "AM_INSTALLATION_DIGEST.txt"),
            )

            # Make sure the setup worked as expected, otherwise the test is meaningless
            self.assertTrue(
                os.path.exists(os.path.join(macro_directory, "FakeMacro1.FCMacro"))
            )
            self.assertTrue(
                os.path.exists(os.path.join(macro_directory, "FakeMacro2.FCMacro"))
            )
            self.assertTrue(
                os.path.exists(os.path.join(macro_directory, "FakeMacro3.FCMacro"))
            )

            self.test_object.remove_extra_files(toplevel_path)  # Shouldn't throw

            self.assertFalse(
                os.path.exists(os.path.join(macro_directory, "FakeMacro1.FCMacro"))
            )
            self.assertFalse(
                os.path.exists(os.path.join(macro_directory, "FakeMacro2.FCMacro"))
            )
            self.assertFalse(
                os.path.exists(os.path.join(macro_directory, "FakeMacro3.FCMacro"))
            )

    def test_runs_uninstaller_script_successful(self):
        """Tests that the uninstall.py script is called"""
        with tempfile.TemporaryDirectory() as temp_dir:
            toplevel_path = self.setup_dummy_installation(temp_dir)
            with open(
                os.path.join(toplevel_path, "uninstall.py"), "w", encoding="utf-8"
            ) as f:
                double_escaped = temp_dir.replace("\\", "\\\\")
                f.write(
                    f"""# Mock uninstaller script
import os
path = '{double_escaped}'
with open(os.path.join(path,"RAN_UNINSTALLER.txt"),"w",encoding="utf-8") as f:
    f.write("File created by uninstall.py from unit tests")
"""
                )
            self.test_object.run_uninstall_script(
                toplevel_path
            )  # The exception does not leak out
            self.assertTrue(
                os.path.exists(os.path.join(temp_dir, "RAN_UNINSTALLER.txt"))
            )

    def test_runs_uninstaller_script_failure(self):
        """Tests that exceptions in the uninstall.py script do not leak out"""
        with tempfile.TemporaryDirectory() as temp_dir:
            toplevel_path = self.setup_dummy_installation(temp_dir)
            with open(
                os.path.join(toplevel_path, "uninstall.py"), "w", encoding="utf-8"
            ) as f:
                f.write(
                    f"""# Mock uninstaller script
raise RuntimeError("Fake exception for unit testing")
"""
                )
            self.test_object.run_uninstall_script(
                toplevel_path
            )  # The exception does not leak out


class TestMacroUninstaller(unittest.TestCase):
    """Test class for addonmanager_uninstaller.py non-GUI functionality"""

    MODULE = "test_uninstaller"  # file name without extension

    def setUp(self):
        self.mock_addon = MockAddon()
        self.mock_addon.macro = MockMacro()
        self.test_object = MacroUninstaller(self.mock_addon)
        self.signals_caught = []

        self.test_object.finished.connect(
            functools.partial(self.catch_signal, "finished")
        )
        self.test_object.success.connect(
            functools.partial(self.catch_signal, "success")
        )
        self.test_object.failure.connect(
            functools.partial(self.catch_signal, "failure")
        )

    def tearDown(self):
        pass

    def catch_signal(self, signal_name, *_):
        """Internal use: used to catch and log any emitted signals. Not called directly."""
        self.signals_caught.append(signal_name)

    def test_remove_simple_macro(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            self.test_object.installation_location = temp_dir
            self.mock_addon.macro.install(temp_dir)
            # Make sure the setup worked, otherwise the test is meaningless
            self.assertTrue(
                os.path.exists(os.path.join(temp_dir, self.mock_addon.macro.filename))
            )
            self.test_object.run()
            self.assertFalse(
                os.path.exists(os.path.join(temp_dir, self.mock_addon.macro.filename))
            )
            self.assertNotIn("failure", self.signals_caught)
            self.assertIn("success", self.signals_caught)
            self.assertIn("finished", self.signals_caught)

    def test_remove_macro_with_icon(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            self.test_object.installation_location = temp_dir
            self.mock_addon.macro.icon = "mock_icon_test.svg"
            self.mock_addon.macro.install(temp_dir)
            # Make sure the setup worked, otherwise the test is meaningless
            self.assertTrue(
                os.path.exists(os.path.join(temp_dir, self.mock_addon.macro.filename))
            )
            self.assertTrue(
                os.path.exists(os.path.join(temp_dir, self.mock_addon.macro.icon))
            )
            self.test_object.run()
            self.assertFalse(
                os.path.exists(os.path.join(temp_dir, self.mock_addon.macro.filename))
            )
            self.assertFalse(
                os.path.exists(os.path.join(temp_dir, self.mock_addon.macro.icon))
            )
            self.assertNotIn("failure", self.signals_caught)
            self.assertIn("success", self.signals_caught)
            self.assertIn("finished", self.signals_caught)

    def test_remove_macro_with_xpm_data(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            self.test_object.installation_location = temp_dir
            self.mock_addon.macro.xpm = "/*Fake XPM data*/"
            self.mock_addon.macro.install(temp_dir)
            # Make sure the setup worked, otherwise the test is meaningless
            self.assertTrue(
                os.path.exists(os.path.join(temp_dir, self.mock_addon.macro.filename))
            )
            self.assertTrue(
                os.path.exists(os.path.join(temp_dir, "MockMacro_icon.xpm"))
            )
            self.test_object.run()
            self.assertFalse(
                os.path.exists(os.path.join(temp_dir, self.mock_addon.macro.filename))
            )
            self.assertFalse(
                os.path.exists(os.path.join(temp_dir, "MockMacro_icon.xpm"))
            )
            self.assertNotIn("failure", self.signals_caught)
            self.assertIn("success", self.signals_caught)
            self.assertIn("finished", self.signals_caught)

    def test_remove_macro_with_files(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            self.test_object.installation_location = temp_dir
            self.mock_addon.macro.other_files = [
                "test_file_1.txt",
                "test_file_2.FCMacro",
                "subdir/test_file_3.txt",
            ]
            self.mock_addon.macro.install(temp_dir)
            # Make sure the setup worked, otherwise the test is meaningless
            for f in self.mock_addon.macro.other_files:
                self.assertTrue(
                    os.path.exists(os.path.join(temp_dir, f)),
                    f"Expected {f} to exist, and it does not",
                )
            self.test_object.run()
            for f in self.mock_addon.macro.other_files:
                self.assertFalse(
                    os.path.exists(os.path.join(temp_dir, f)),
                    f"Expected {f} to be removed, and it was not",
                )
            self.assertFalse(
                os.path.exists(os.path.join(temp_dir, "subdir")),
                "Failed to remove empty subdirectory",
            )
            self.assertNotIn("failure", self.signals_caught)
            self.assertIn("success", self.signals_caught)
            self.assertIn("finished", self.signals_caught)

    def test_remove_nonexistent_macro(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            self.test_object.installation_location = temp_dir
            # Don't run the installer:

            self.assertFalse(
                os.path.exists(os.path.join(temp_dir, self.mock_addon.macro.filename))
            )
            self.test_object.run()  # Should not raise an exception
            self.assertFalse(
                os.path.exists(os.path.join(temp_dir, self.mock_addon.macro.filename))
            )
            self.assertNotIn("failure", self.signals_caught)
            self.assertIn("success", self.signals_caught)
            self.assertIn("finished", self.signals_caught)

    def test_remove_write_protected_macro(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            self.test_object.installation_location = temp_dir
            self.mock_addon.macro.install(temp_dir)
            # Make sure the setup worked, otherwise the test is meaningless
            f = os.path.join(temp_dir, self.mock_addon.macro.filename)
            self.assertTrue(os.path.exists(f))
            os.chmod(f, S_IREAD | S_IRGRP | S_IROTH)
            self.test_object.run()

            if os.path.exists(f):
                os.chmod(f, S_IWUSR | S_IREAD)
                self.assertNotIn("success", self.signals_caught)
                self.assertIn("failure", self.signals_caught)
            else:
                # In some cases we managed to delete it anyway:
                self.assertIn("success", self.signals_caught)
                self.assertNotIn("failure", self.signals_caught)
            self.assertIn("finished", self.signals_caught)

    def test_cleanup_directories_multiple_empty(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            empty_directories = set(["empty1", "empty2", "empty3"])
            full_paths = set()
            for directory in empty_directories:
                full_path = os.path.join(temp_dir, directory)
                os.mkdir(full_path)
                full_paths.add(full_path)

            for directory in full_paths:
                self.assertTrue(directory, "Test code failed to create {directory}")
            self.test_object._cleanup_directories(full_paths)
            for directory in full_paths:
                self.assertFalse(os.path.exists(directory))

    def test_cleanup_directories_none(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            full_paths = set()
            self.test_object._cleanup_directories(full_paths)  # Shouldn't throw

    def test_cleanup_directories_not_empty(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            empty_directories = set(["empty1", "empty2", "empty3"])
            full_paths = set()
            for directory in empty_directories:
                full_path = os.path.join(temp_dir, directory)
                os.mkdir(full_path)
                full_paths.add(full_path)
                with open(
                    os.path.join(full_path, "test.txt"), "w", encoding="utf-8"
                ) as f:
                    f.write("Unit test dummy data\n")

            for directory in full_paths:
                self.assertTrue(directory, "Test code failed to create {directory}")
            self.test_object._cleanup_directories(full_paths)
            for directory in full_paths:
                self.assertTrue(os.path.exists(directory))



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022-2023 FreeCAD Project Association                   *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

import addonmanager_freecad_interface as fci

# Unit test for the Addon Manager module GUI
from AddonManagerTest.gui.test_gui import TestGui as AddonManagerTestGui

from AddonManagerTest.gui.test_workers_utility import (
    TestWorkersUtility as AddonManagerTestWorkersUtility,
)
from AddonManagerTest.gui.test_workers_startup import (
    TestWorkersStartup as AddonManagerTestWorkersStartup,
)
from AddonManagerTest.gui.test_installer_gui import (
    TestInstallerGui as AddonManagerTestInstallerGui,
)
from AddonManagerTest.gui.test_installer_gui import (
    TestMacroInstallerGui as AddonManagerTestMacroInstallerGui,
)
from AddonManagerTest.gui.test_update_all_gui import (
    TestUpdateAllGui as AddonManagerTestUpdateAllGui,
)
from AddonManagerTest.gui.test_uninstaller_gui import (
    TestUninstallerGUI as AddonManagerTestUninstallerGUI,
)


class TestListTerminator:
    pass


# Basic usage mostly to get static analyzers to stop complaining about unused imports
loaded_gui_tests = [
    AddonManagerTestGui,
    AddonManagerTestWorkersUtility,
    AddonManagerTestWorkersStartup,
    AddonManagerTestInstallerGui,
    AddonManagerTestMacroInstallerGui,
    AddonManagerTestUpdateAllGui,
    AddonManagerTestUninstallerGUI,
    TestListTerminator,  # Needed to prevent the last test from running twice
]
for test in loaded_gui_tests:
    fci.Console.PrintLog(f"Loaded tests from {test.__name__}\n")



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

""" System for checking the network connection status asynchronously. """

import FreeCAD

from PySide import QtCore, QtWidgets

import NetworkManager
from addonmanager_workers_utility import ConnectionChecker

translate = FreeCAD.Qt.translate


class ConnectionCheckerGUI(QtCore.QObject):
    """Determine whether there is an active network connection, showing a progress message if it
    starts to take too long, and an error message if the network cannot be accessed."""

    connection_available = QtCore.Signal()
    check_complete = QtCore.Signal()

    def __init__(self):
        super().__init__()

        # Check the connection in a new thread, so FreeCAD stays responsive
        self.connection_checker = ConnectionChecker()
        self.connection_checker.success.connect(self._check_succeeded)
        self.connection_checker.failure.connect(self._network_connection_failed)

        self.connection_message_timer = None
        self.connection_check_message = None

    def start(self):
        """Start the connection check"""
        self.connection_checker.start()

        # If it takes longer than a half second to check the connection, show a message:
        self.connection_message_timer = QtCore.QTimer.singleShot(
            500, self._show_connection_check_message
        )

    def _show_connection_check_message(self):
        """Display a message informing the user that the check is in process"""
        if not self.connection_checker.isFinished():
            self.connection_check_message = QtWidgets.QMessageBox(
                QtWidgets.QMessageBox.Information,
                translate("AddonsInstaller", "Checking connection"),
                translate("AddonsInstaller", "Checking for connection to GitHub..."),
                QtWidgets.QMessageBox.Cancel,
            )
            self.connection_check_message.buttonClicked.connect(
                self.cancel_network_check
            )
            self.connection_check_message.show()

    def cancel_network_check(self, _):
        """Cancel the check"""
        if not self.connection_checker.isFinished():
            self.connection_checker.success.disconnect(self._check_succeeded)
            self.connection_checker.failure.disconnect(self._network_connection_failed)
            self.connection_checker.requestInterruption()
            self.connection_checker.wait(500)
            self.connection_check_message.close()
            self.check_complete.emit()

    def _network_connection_failed(self, message: str) -> None:
        """Callback for failed connection check. Displays an error message, then emits the
        check_complete signal (but not the connection available signal)."""
        # This must run on the main GUI thread
        if hasattr(self, "connection_check_message") and self.connection_check_message:
            self.connection_check_message.close()
        if NetworkManager.HAVE_QTNETWORK:
            QtWidgets.QMessageBox.critical(
                None, translate("AddonsInstaller", "Connection failed"), message
            )
        else:
            # pylint: disable=line-too-long
            QtWidgets.QMessageBox.critical(
                None,
                translate("AddonsInstaller", "Missing dependency"),
                translate(
                    "AddonsInstaller",
                    "Could not import QtNetwork -- see Report View for details. Addon Manager unavailable.",
                ),
            )

        self.check_complete.emit()

    def _check_succeeded(self):
        """Emit both the connection_available and the check_complete signals, in that order."""

        if hasattr(self, "connection_check_message") and self.connection_check_message:
            self.connection_check_message.close()

        self.connection_available.emit()
        self.check_complete.emit()



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022-2023 FreeCAD Project Association                   *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""Tests for the MacroParser class"""

import io
import os
import sys
import unittest

sys.path.append("../../")  # So the IDE can find the classes to run with

from addonmanager_macro_parser import MacroParser
from AddonManagerTest.app.mocks import MockConsole, CallCatcher, MockThread


# pylint: disable=protected-access, too-many-public-methods


class TestMacroParser(unittest.TestCase):
    """Test the MacroParser class"""

    def setUp(self) -> None:
        self.test_object = MacroParser("UnitTestMacro")
        self.test_object.console = MockConsole()
        self.test_object.current_thread = MockThread()

    def tearDown(self) -> None:
        pass

    def test_fill_details_from_code_normal(self):
        """Test to make sure _process_line gets called as expected"""
        catcher = CallCatcher()
        self.test_object._process_line = catcher.catch_call
        fake_macro_data = self.given_some_lines(20, 10)
        self.test_object.fill_details_from_code(fake_macro_data)
        self.assertEqual(catcher.call_count, 10)

    def test_fill_details_from_code_too_many_lines(self):
        """Test to make sure _process_line gets limited as expected"""
        catcher = CallCatcher()
        self.test_object._process_line = catcher.catch_call
        self.test_object.MAX_LINES_TO_SEARCH = 5
        fake_macro_data = self.given_some_lines(20, 10)
        self.test_object.fill_details_from_code(fake_macro_data)
        self.assertEqual(catcher.call_count, 5)

    def test_fill_details_from_code_thread_interrupted(self):
        """Test to make sure _process_line gets stopped as expected"""
        catcher = CallCatcher()
        self.test_object._process_line = catcher.catch_call
        self.test_object.current_thread.interrupt_after_n_calls = 6  # Stop on the 6th
        fake_macro_data = self.given_some_lines(20, 10)
        self.test_object.fill_details_from_code(fake_macro_data)
        self.assertEqual(catcher.call_count, 5)

    @staticmethod
    def given_some_lines(num_lines, num_dunder_lines) -> str:
        """Generate fake macro header data with the given number of lines and number of
        lines beginning with a double-underscore."""
        result = ""
        for i in range(num_lines):
            if i < num_dunder_lines:
                result += f"__something_{i}__ = 'Test{i}'  # A line to be scanned\n"
            else:
                result += f"# Nothing to see on line {i}\n"
        return result

    def test_process_line_known_lines(self):
        """Lines starting with keys are processed"""
        test_lines = ["__known_key__ = 'Test'", "__another_known_key__ = 'Test'"]
        for line in test_lines:
            with self.subTest(line=line):
                self.test_object.remaining_item_map = {
                    "__known_key__": "known_key",
                    "__another_known_key__": "another_known_key",
                }
                content_lines = io.StringIO(line)
                read_in_line = content_lines.readline()
                catcher = CallCatcher()
                self.test_object._process_key = catcher.catch_call
                self.test_object._process_line(read_in_line, content_lines)
                self.assertTrue(
                    catcher.called, "_process_key was not called for a known key"
                )

    def test_process_line_unknown_lines(self):
        """Lines starting with non-keys are not processed"""
        test_lines = [
            "# Just a line with a comment",
            "\n",
            "__dont_know_this_one__ = 'Who cares?'",
            "# __known_key__ = 'Aha, but it is commented out!'",
        ]
        for line in test_lines:
            with self.subTest(line=line):
                self.test_object.remaining_item_map = {
                    "__known_key__": "known_key",
                    "__another_known_key__": "another_known_key",
                }
                content_lines = io.StringIO(line)
                read_in_line = content_lines.readline()
                catcher = CallCatcher()
                self.test_object._process_key = catcher.catch_call
                self.test_object._process_line(read_in_line, content_lines)
                self.assertFalse(
                    catcher.called, "_process_key was called for an unknown key"
                )

    def test_process_key_standard(self):
        """Normal expected data is processed"""
        self.test_object._reset_map()
        in_memory_data = '__comment__ = "Test"'
        content_lines = io.StringIO(in_memory_data)
        line = content_lines.readline()
        self.test_object._process_key("__comment__", line, content_lines)
        self.assertTrue(self.test_object.parse_results["comment"], "Test")

    def test_process_key_special(self):
        """Special handling for version = date is processed"""
        self.test_object._reset_map()
        self.test_object.parse_results["date"] = "2001-01-01"
        in_memory_data = "__version__ = __date__"
        content_lines = io.StringIO(in_memory_data)
        line = content_lines.readline()
        self.test_object._process_key("__version__", line, content_lines)
        self.assertTrue(self.test_object.parse_results["version"], "2001-01-01")

    def test_handle_backslash_continuation_no_backslashes(self):
        """The backslash handling code doesn't change a line with no backslashes"""
        in_memory_data = '"Not a backslash in sight"'
        content_lines = io.StringIO(in_memory_data)
        line = content_lines.readline()
        result = self.test_object._handle_backslash_continuation(line, content_lines)
        self.assertEqual(result, in_memory_data)

    def test_handle_backslash_continuation(self):
        """Lines ending in a backslash get stripped and concatenated"""
        in_memory_data = '"Line1\\\nLine2\\\nLine3\\\nLine4"'
        content_lines = io.StringIO(in_memory_data)
        line = content_lines.readline()
        result = self.test_object._handle_backslash_continuation(line, content_lines)
        self.assertEqual(result, '"Line1Line2Line3Line4"')

    def test_handle_triple_quoted_string_no_triple_quotes(self):
        """The triple-quote handler leaves alone lines without triple-quotes"""
        in_memory_data = '"Line1"'
        content_lines = io.StringIO(in_memory_data)
        line = content_lines.readline()
        result, was_triple_quoted = self.test_object._handle_triple_quoted_string(
            line, content_lines
        )
        self.assertEqual(result, in_memory_data)
        self.assertFalse(was_triple_quoted)

    def test_handle_triple_quoted_string(self):
        """Data is extracted across multiple lines for a triple-quoted string"""
        in_memory_data = '"""Line1\nLine2\nLine3\nLine4"""\nLine5\n'
        content_lines = io.StringIO(in_memory_data)
        line = content_lines.readline()
        result, was_triple_quoted = self.test_object._handle_triple_quoted_string(
            line, content_lines
        )
        self.assertEqual(result, '"""Line1\nLine2\nLine3\nLine4"""')
        self.assertTrue(was_triple_quoted)

    def test_strip_quotes_single(self):
        """Single quotes are stripped from the final string"""
        expected = "test"
        quoted = f"'{expected}'"
        actual = self.test_object._strip_quotes(quoted)
        self.assertEqual(actual, expected)

    def test_strip_quotes_double(self):
        """Double quotes are stripped from the final string"""
        expected = "test"
        quoted = f'"{expected}"'
        actual = self.test_object._strip_quotes(quoted)
        self.assertEqual(actual, expected)

    def test_strip_quotes_triple(self):
        """Triple quotes are stripped from the final string"""
        expected = "test"
        quoted = f'"""{expected}"""'
        actual = self.test_object._strip_quotes(quoted)
        self.assertEqual(actual, expected)

    def test_strip_quotes_unquoted(self):
        """Unquoted data results in None"""
        unquoted = "This has no quotation marks of any kind"
        actual = self.test_object._strip_quotes(unquoted)
        self.assertIsNone(actual)

    def test_standard_extraction_string(self):
        """String variables are extracted and stored"""
        string_keys = [
            "comment",
            "url",
            "wiki",
            "version",
            "author",
            "date",
            "icon",
            "xpm",
        ]
        for key in string_keys:
            with self.subTest(key=key):
                self.test_object._standard_extraction(key, "test")
                self.assertEqual(self.test_object.parse_results[key], "test")

    def test_standard_extraction_list(self):
        """List variable is extracted and stored"""
        key = "other_files"
        self.test_object._standard_extraction(key, "test1, test2, test3")
        self.assertIn("test1", self.test_object.parse_results[key])
        self.assertIn("test2", self.test_object.parse_results[key])
        self.assertIn("test3", self.test_object.parse_results[key])

    def test_apply_special_handling_version(self):
        """If the tag is __version__, apply our special handling"""
        self.test_object._reset_map()
        self.test_object._apply_special_handling("__version__", 42)
        self.assertNotIn("__version__", self.test_object.remaining_item_map)
        self.assertEqual(self.test_object.parse_results["version"], "42")

    def test_apply_special_handling_not_version(self):
        """If the tag is not __version__, raise an error"""
        self.test_object._reset_map()
        with self.assertRaises(SyntaxError):
            self.test_object._apply_special_handling("__not_version__", 42)
        self.assertIn("__version__", self.test_object.remaining_item_map)

    def test_process_noncompliant_version_date(self):
        """Detect and allow __date__ for the __version__"""
        self.test_object.parse_results["date"] = "1/2/3"
        self.test_object._process_noncompliant_version("__date__")
        self.assertEqual(
            self.test_object.parse_results["version"],
            self.test_object.parse_results["date"],
        )

    def test_process_noncompliant_version_float(self):
        """Detect and allow floats for the __version__"""
        self.test_object._process_noncompliant_version(1.2)
        self.assertEqual(self.test_object.parse_results["version"], "1.2")

    def test_process_noncompliant_version_int(self):
        """Detect and allow integers for the __version__"""
        self.test_object._process_noncompliant_version(42)
        self.assertEqual(self.test_object.parse_results["version"], "42")

    def test_detect_illegal_content_prefixed_string(self):
        """Detect and raise an error for various kinds of prefixed strings"""
        illegal_strings = [
            "f'Some fancy {thing}'",
            'f"Some fancy {thing}"',
            "r'Some fancy {thing}'",
            'r"Some fancy {thing}"',
            "u'Some fancy {thing}'",
            'u"Some fancy {thing}"',
            "fr'Some fancy {thing}'",
            'fr"Some fancy {thing}"',
            "rf'Some fancy {thing}'",
            'rf"Some fancy {thing}"',
        ]
        for test_string in illegal_strings:
            with self.subTest(test_string=test_string):
                with self.assertRaises(SyntaxError):
                    MacroParser._detect_illegal_content(test_string)

    def test_detect_illegal_content_not_a_string(self):
        """Detect and raise an error for (some) non-strings"""
        illegal_strings = [
            "no quotes",
            "do_stuff()",
            'print("A function call sporting quotes!")',
            "__name__",
            "__version__",
            "1.2.3",
        ]
        for test_string in illegal_strings:
            with self.subTest(test_string=test_string):
                with self.assertRaises(SyntaxError):
                    MacroParser._detect_illegal_content(test_string)

    def test_detect_illegal_content_no_failure(self):
        """Recognize strings of various kinds, plus ints, and floats"""
        legal_strings = [
            '"Some legal value in double quotes"',
            "'Some legal value in single quotes'",
            '"""Some legal value in triple quotes"""',
            "__date__",
            "42",
            "4.2",
        ]
        for test_string in legal_strings:
            with self.subTest(test_string=test_string):
                MacroParser._detect_illegal_content(test_string)

    #####################
    # INTEGRATION TESTS #
    #####################

    def test_macro_parser(self):
        """INTEGRATION TEST: Given "real" data, ensure the parsing yields the expected results."""
        data_dir = os.path.join(os.path.dirname(__file__), "../data")
        macro_file = os.path.join(data_dir, "DoNothing.FCMacro")
        with open(macro_file, "r", encoding="utf-8") as f:
            code = f.read()
        self.test_object.fill_details_from_code(code)
        self.assertEqual(len(self.test_object.console.errors), 0)
        self.assertEqual(len(self.test_object.console.warnings), 0)
        self.assertEqual(self.test_object.parse_results["author"], "Chris Hennes")
        self.assertEqual(self.test_object.parse_results["version"], "1.0")
        self.assertEqual(self.test_object.parse_results["date"], "2022-02-28")
        self.assertEqual(
            self.test_object.parse_results["comment"],
            "Do absolutely nothing. For Addon Manager integration tests.",
        )
        self.assertEqual(
            self.test_object.parse_results["url"], "https://github.com/FreeCAD/FreeCAD"
        )
        self.assertEqual(self.test_object.parse_results["icon"], "not_real.png")
        self.assertListEqual(
            self.test_object.parse_results["other_files"],
            ["file1.py", "file2.py", "file3.py"],
        )
        self.assertNotEqual(self.test_object.parse_results["xpm"], "")



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

""" Contains a wrapper class for a table listing authors and maintainers """

import os

from PySide.QtWidgets import QTableWidgetItem
from PySide.QtGui import QIcon
from PySide.QtCore import Qt

import FreeCAD
import FreeCADGui

from addonmanager_devmode_person_editor import PersonEditor

translate = FreeCAD.Qt.translate

# pylint: disable=too-few-public-methods


class PeopleTable:
    """A QTableWidget and associated buttons for managing the list of authors and maintainers."""

    def __init__(self):
        self.widget = FreeCADGui.PySideUic.loadUi(
            os.path.join(os.path.dirname(__file__), "developer_mode_people_table.ui")
        )

        self.widget.addButton.setIcon(
            QIcon.fromTheme("add", QIcon(":/icons/list-add.svg"))
        )
        self.widget.removeButton.setIcon(
            QIcon.fromTheme("remove", QIcon(":/icons/list-remove.svg"))
        )

        self.widget.addButton.clicked.connect(self._add_clicked)
        self.widget.removeButton.clicked.connect(self._remove_clicked)
        self.widget.tableWidget.itemSelectionChanged.connect(self._selection_changed)
        self.widget.tableWidget.itemDoubleClicked.connect(self._edit)
        self.metadata = None

    def show(self, metadata):
        """Set up the widget based on incoming metadata"""
        self.metadata = metadata
        self._populate_from_metadata()
        self.widget.removeButton.setDisabled(True)
        self.widget.show()

    def _populate_from_metadata(self):
        """Use the passed metadata object to populate the maintainers and authors"""
        self.widget.tableWidget.setRowCount(0)
        row = 0
        for maintainer in self.metadata.Maintainer:
            name = maintainer["name"]
            email = maintainer["email"]
            self._add_row(row, "maintainer", name, email)
            row += 1
        for author in self.metadata.Author:
            name = author["name"]
            email = author["email"]
            self._add_row(row, "author", name, email)
            row += 1

    def _add_row(self, row, person_type, name, email):
        """Add this person to the tableWidget at row given"""
        person_type_translation = {
            "maintainer": translate("AddonsInstaller", "Maintainer"),
            "author": translate("AddonsInstaller", "Author"),
        }
        self.widget.tableWidget.insertRow(row)
        item = QTableWidgetItem(person_type_translation[person_type])
        item.setData(Qt.UserRole, person_type)
        self.widget.tableWidget.setItem(row, 0, item)
        self.widget.tableWidget.setItem(row, 1, QTableWidgetItem(name))
        self.widget.tableWidget.setItem(row, 2, QTableWidgetItem(email))

    def _add_clicked(self):
        """Callback: the Add Person button was clicked"""
        dlg = PersonEditor()
        person_type, name, email = dlg.exec()
        if person_type and name:
            self._add_row(self.widget.tableWidget.rowCount(), person_type, name, email)
            if person_type == "maintainer":
                self.metadata.addMaintainer(name, email)
            else:
                self.metadata.addAuthor(name, email)

    def _remove_clicked(self):
        """Callback: the Remove Person button was clicked"""
        items = self.widget.tableWidget.selectedIndexes()
        if items:
            # We only support single-selection, so can just pull the row # from
            # the first entry
            row = items[0].row()
            person_type = self.widget.tableWidget.item(row, 0).data(Qt.UserRole)
            name = self.widget.tableWidget.item(row, 1).text()
            email = self.widget.tableWidget.item(row, 2).text()
            self.widget.tableWidget.removeRow(row)
            if person_type == "maintainer":
                self.metadata.removeMaintainer(name, email)
            else:
                self.metadata.removeAuthor(name, email)

    def _edit(self, item):
        """Callback: a row in the tableWidget was double-clicked"""
        row = item.row()
        person_type = self.widget.tableWidget.item(row, 0).data(Qt.UserRole)
        name = self.widget.tableWidget.item(row, 1).text()
        email = self.widget.tableWidget.item(row, 2).text()

        dlg = PersonEditor()
        dlg.setup(person_type, name, email)
        new_person_type, new_name, new_email = dlg.exec()

        if new_person_type and new_name:
            self.widget.tableWidget.removeRow(row)
            if person_type == "maintainer":
                self.metadata.removeMaintainer(name, email)
            else:
                self.metadata.removeAuthor(name, email)
            self._add_row(row, new_person_type, new_name, email)
            if new_person_type == "maintainer":
                self.metadata.addMaintainer(new_name, new_email)
            else:
                self.metadata.addAuthor(new_name, new_email)
            self.widget.tableWidget.selectRow(row)

    def _selection_changed(self):
        """Callback: the current selection in the tableWidget changed"""
        items = self.widget.tableWidget.selectedItems()
        if items:
            self.widget.removeButton.setDisabled(False)
        else:
            self.widget.removeButton.setDisabled(True)



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""Wrap QtCore imports so that can be replaced when running outside of FreeCAD (e.g. for
unit tests, etc.) Only provides wrappers for the things commonly used by the Addon
Manager."""

try:
    from PySide import QtCore

    QObject = QtCore.QObject
    Signal = QtCore.Signal

    def is_interruption_requested() -> bool:
        return QtCore.QThread.currentThread().isInterruptionRequested()

except ImportError:
    QObject = object

    class Signal:
        """A purely synchronous signal. emit() does not use queued slots so cannot be
        used across threads."""

        def __init__(self, *args):
            self.expected_types = args
            self.connections = []

        def connect(self, func):
            self.connections.append(func)

        def disconnect(self, func):
            if func in self.connections:
                self.connections.remove(func)

        def emit(self, *args):
            for connection in self.connections:
                connection(args)

    def is_interruption_requested() -> bool:
        return False
